/* 
   List.h - templated list class

   Copyright (C) Xi 18, 1999.

   This file is part of blackhome.

   blackhome is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   blackhome is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with blackhome; see the file COPYING.  If not, write to the Free
   Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

 /* ******************************************************************************************************************* **
 **                                                                                                                     **
 **  XeeEngine vertex description                                                                                       **
 **                                                                                                                     **
 **  File dates: Earliest [5th January 1999] Most Recent [14 October 1999]                                              **
 **                                                                                                                     **
 **  - Templated doubly linked list capable of keeping a list of anything (very specific to my needs !!!)               **
 **  - Visual C++ 5.0 only understands templated classes if both header and code are in a header file !!!               **
 **                                                                                                                     **
 ** ******************************************************************************************************************* */
 

#ifndef __LISTH__TEMPLATED_DOUBLY_LINKED_LIST
#define __LISTH__TEMPLATED_DOUBLY_LINKED_LIST


// includes
#include "errno.h"


// prototypes
template<class MyType> class List;


/* ********************************************************************************************************* **
ListItem: This class is a list node. It stores an object of templated type and has forward and backward
          links to other list nodes, making it a doubly linked list.
** ********************************************************************************************************* */
template<class MyType>
struct ListItem
{
	public:
		ListItem(MyType &,ListItem<MyType> *,ListItem<MyType> *);

		MyType & Get(void)    { return _data;  }	// provides read/write access to _data
		ListItem * Prev(void) { return _pPrev; }	// provides read access to _pPrev
		ListItem * Next(void) { return _pNext; }	// provides read access to _pNext

	protected:
		ListItem * _pPrev;
		ListItem * _pNext;
		MyType _data;

		friend class List<MyType>;
};

/* ********************************************************************************************************* **
List: The list class.
** ********************************************************************************************************* */
template<class MyType>
class List
{
	public:
		List();
		List(List<MyType> &);
		~List();

		void Add(MyType &);								// append a node to the tail of the list
		void AddHead(MyType &);							// append a node to the head of the list
		void InsertAfter(MyType &,ListItem<MyType> *);	// insert node after a given node
		void InsertBefore(MyType &,ListItem<MyType> *);	// insert node before a given node
		void Remove(ListItem<MyType> *);				// remove a node from any point in the list
		void Remove(MyType &);							// remove nodes containing the given data
		void RemoveItemAt(long);						// remove nth item (indices begin at zero)
		void Release(void);								// release the entire list
		long Count(void);								// how many items are in the list ?
		void Append(List<MyType> &);					// append the contents of the given list to this one
		void SetErrorObject(MyType &);
		void SetDeleteFunc(void (*)(MyType &));

		ListItem<MyType> * GetNodeAt(long);
		ListItem<MyType> * GetHead(void);							// get a pointer to the first node in the list
		ListItem<MyType> * GetTail(void);							// get a pointer to the last node in the list
		ListItem<MyType> * Find(MyType &,ListItem<MyType> * = 0);	// find a node with a matching data item

		List<MyType> & operator = (List<MyType> &);		// copy the contents of a list into this one
		MyType &       operator [] (long);

	protected:
		ListItem<MyType> * _pHead;
		ListItem<MyType> * _pTail;
		ListItem<MyType> * _pLastUsedItem;
		MyType _errorObject;
		void (* _deleteFunc)(MyType &);
		long _count;
		long _lastUsedIndex;
		bool _fCountNeedsUpdating;
		bool _fRandomAccessIndexNeedsUpdating;

		void CopyList(List &);
};


/* ********************************************************************************************************* **
Constructor (LIST): Reset state.
** ********************************************************************************************************* */
template<class MyType>
List<MyType>::List()
{
	_pHead = 0;
	_pTail = 0;
	_deleteFunc = 0;
	_fCountNeedsUpdating = true;
	_fRandomAccessIndexNeedsUpdating = true;
	_lastUsedIndex = 0;
	_count = 0;
}

template<class MyType>
List<MyType>::List(List<MyType> & source)
{
	_pHead = 0;
	_pTail = 0;
	_deleteFunc = 0;
	_fCountNeedsUpdating = true;
	_fRandomAccessIndexNeedsUpdating = true;
	_lastUsedIndex = 0;
	_count = 0;
	CopyList(source);
}

/* ********************************************************************************************************* **
Destructor (LIST): Release everything.
** ********************************************************************************************************* */
template<class MyType>
List<MyType>::~List()
{
	Release();
}

/* ********************************************************************************************************* **
Release (LIST): Destroy all nodes, but not their data.
** ********************************************************************************************************* */
template<class MyType>
void List<MyType>::Release(void)
{
	ListItem<MyType> * pNext;
	ListItem<MyType> * pVictim;
	
	pVictim = _pHead;
	while(pVictim)
	{
		pNext = pVictim->Next();

		if(_deleteFunc)
			_deleteFunc(pVictim->Get());

		delete pVictim;
		pVictim = pNext;
	}

	_pHead = 0;
	_pTail = 0;
	_fCountNeedsUpdating = true;
	_fRandomAccessIndexNeedsUpdating = true;
}

/* ********************************************************************************************************* **
Remove (LIST): Unlink a individual node from the list.
** ********************************************************************************************************* */
template<class MyType>
void List<MyType>::Remove(ListItem<MyType> * pVictim)
{
	if(_deleteFunc)
		_deleteFunc(pVictim->Get());

	if(_pHead == pVictim) _pHead = _pHead->_pNext;
	if(_pTail == pVictim) _pTail = _pTail->_pPrev;
	if(pVictim->_pNext) pVictim->_pNext->_pPrev = pVictim->_pPrev;
	if(pVictim->_pPrev) pVictim->_pPrev->_pNext = pVictim->_pNext;
	delete pVictim;

	_fCountNeedsUpdating = true;
	_fRandomAccessIndexNeedsUpdating = true;
}

/* ********************************************************************************************************* **
Remove (LIST): Unlink node at the given index in the list (indices begin at zero)
** ********************************************************************************************************* */
template<class MyType>
void List<MyType>::RemoveItemAt(long index)
{
	ListItem<MyType> * pNext;
	ListItem<MyType> * pVictim;
	long i = 0;

	pVictim = _pHead;
	while(pVictim)
	{
		pNext = pVictim->Next();

		if(i++ == index)
			Remove(pVictim);

		pVictim = pNext;
	}

	_fCountNeedsUpdating = true;
	_fRandomAccessIndexNeedsUpdating = true;
}

/* ********************************************************************************************************* **
Remove (LIST): Unlink all nodes containing the given data
** ********************************************************************************************************* */
template<class MyType>
void List<MyType>::Remove(MyType & data)
{
	ListItem<MyType> * pNext;
	ListItem<MyType> * pVictim;

	pVictim = _pHead;
	while(pVictim)
	{
		pNext = pVictim->Next();

		if(pVictim->_data == data)
			Remove(pVictim);

		pVictim = pNext;
	}

	_fCountNeedsUpdating = true;
	_fRandomAccessIndexNeedsUpdating = true;
}

/* ********************************************************************************************************* **
Add (LIST): Add a new node to the tail of the list.
** ********************************************************************************************************* */
template<class MyType>
void List<MyType>::Add(MyType & data)
{
	InsertAfter(data,_pTail);
}

/* ********************************************************************************************************* **
AddHead (LIST): Add a new node to the head of the list.
** ********************************************************************************************************* */
template<class MyType>
void List<MyType>::AddHead(MyType & data)
{
	InsertBefore(data,_pHead);
}

/* ********************************************************************************************************* **
InsertAfter (LIST): Insert a node after a given node.
** ********************************************************************************************************* */
template<class MyType>
void List<MyType>::InsertAfter(MyType & data,ListItem<MyType> * pListItem)
{
	if(pListItem)
	{
		ListItem<MyType> * pNextListItem = pListItem->_pNext;
		pListItem->_pNext = new ListItem<MyType> (data,pListItem,pNextListItem);

		if(pListItem == _pTail)
			_pTail = _pTail->_pNext;
		else
			pNextListItem->_pPrev = pListItem->_pNext;
	}
	else
		_pHead = _pTail = new ListItem<MyType> (data,0,0);

	_fCountNeedsUpdating = true;
	_fRandomAccessIndexNeedsUpdating = true;
}

/* ********************************************************************************************************* **
InsertBefore (LIST): Insert a node before a given node.
** ********************************************************************************************************* */
template<class MyType>
void List<MyType>::InsertBefore(MyType & data,ListItem<MyType> * pListItem)
{
	if(pListItem)
	{
		ListItem<MyType> * pPrevListItem = pListItem->_pPrev;
		pListItem->_pPrev = new ListItem<MyType> (data,pPrevListItem,pListItem);

		if(pListItem == _pHead)
			_pHead = _pHead->_pPrev;
		else
			pPrevListItem->_pNext = pListItem->_pPrev;
	}
	else
		_pHead = _pTail = new ListItem<MyType> (data,0,0);

	_fCountNeedsUpdating = true;
	_fRandomAccessIndexNeedsUpdating = true;
}

/* ********************************************************************************************************* **
Count (LIST): Count how many items are stored in the list.
** ********************************************************************************************************* */
template<class MyType>
long List<MyType>::Count(void)
{
	ListItem<MyType> * pListItem;

	if(_fCountNeedsUpdating)
	{
		_count = 0;
		pListItem = _pHead;

		while(pListItem)
		{
			_count++;
			pListItem = pListItem->Next();
		}

		_fCountNeedsUpdating = false;
	}

	return _count;
}

/* ********************************************************************************************************* **
GetHead (LIST): Get a pointer to the first item in the list.
** ********************************************************************************************************* */
template<class MyType>
ListItem<MyType> * List<MyType>::GetHead(void)
{
	return _pHead;
}

/* ********************************************************************************************************* **
GetTail (LIST): Get a pointer to the last item in the list.
** ********************************************************************************************************* */
template<class MyType>
ListItem<MyType> * List<MyType>::GetTail(void)
{
	return _pTail;
}

/* ********************************************************************************************************* **
Find (LIST): Find a node in the list with the given data stored in it. 
** ********************************************************************************************************* */
template<class MyType>
ListItem<MyType> * List<MyType>::Find(MyType & searchValue,ListItem<MyType> * pFirstListItem)
{
	ListItem<MyType> * pListItem;

	if(pFirstListItem == 0)
		pFirstListItem = _pHead;

	pListItem = pFirstListItem;
	while(pListItem)
	{
		if(pListItem->Get() == searchValue)
			return pListItem;

		pListItem = pListItem->Next();
	}

	return 0;
}

/* ********************************************************************************************************* **
Append (LIST): Append all the nodes from a given list to the tail of this list.
** ********************************************************************************************************* */
template<class MyType>
void List<MyType>::Append(List<MyType> & list)
{
	ListItem<MyType> * pListItem;

	pListItem = list.GetHead();
	while(pListItem)
	{
		Add(pListItem->Get());
		pListItem = pListItem->Next();
	}
}

/* ********************************************************************************************************* **
Operator = (LIST): Release this list, and then add all the nodes from the given list to this one, ie. copy
                   the entire list.
** ********************************************************************************************************* */
template<class MyType>
List<MyType> & List<MyType>::operator = (List<MyType> & source)
{
	CopyList(source);
	return *this;
}

template<class MyType>
void List<MyType>::CopyList(List & source)
{
	Release();

	ListItem<MyType> * pSource;

	pSource = source.GetHead();
	while(pSource)
	{
		Add(pSource->Get());
		pSource = pSource->Next();
	}
}

/* ********************************************************************************************************* **
SetErrorObject (LIST):
** ********************************************************************************************************* */
template<class MyType>
void List<MyType>::SetErrorObject(MyType & t)
{
	_errorObject = t;
}

/* ********************************************************************************************************* **
SetDeleteFunc (LIST):
** ********************************************************************************************************* */
template<class MyType>
void List<MyType>::SetDeleteFunc(void (* deleteFunc)(MyType &))
{
	_deleteFunc = deleteFunc;
}

/* ********************************************************************************************************* **
Operator [] (LIST):
** ********************************************************************************************************* */
template<class MyType>
MyType & List<MyType>::operator [] (long index)
{
	ListItem<MyType> * pItem = GetNodeAt(index);

	if(pItem)
		return pItem->Get();
	
	return _errorObject;
}

template<class MyType>
ListItem<MyType> * List<MyType>::GetNodeAt(long index)
{
	if(index < 0 || index >= Count())
		return 0;
	else
	{
		ListItem<MyType> * pListItem;
		long count;

		if(_fRandomAccessIndexNeedsUpdating)
		{
			pListItem = _pHead;
			count = 0;

			while(pListItem)
			{
				if(count == index)
				{
					_lastUsedIndex = index;
					_pLastUsedItem = pListItem;
					_fRandomAccessIndexNeedsUpdating = false;
					return pListItem;
				}

				count++;
				pListItem = pListItem->Next();
			}
		}
		else
		{
			pListItem = _pLastUsedItem;

			if(index < _lastUsedIndex)
			{
				for(count = _lastUsedIndex; count > index; count--)
				{
					if(pListItem)
						pListItem = pListItem->Prev();
				}
			}
			else
			{
				for(count = _lastUsedIndex; count < index; count++)
				{
					if(pListItem)
						pListItem = pListItem->Next();
				}
			}

			if(pListItem)
			{
				_lastUsedIndex = index;
				_pLastUsedItem = pListItem;
				return pListItem;
			}
		}
	}

	// in theory this line of code cannot be reached
	// unless the list is broken somehow.
	return 0;
}

/* ********************************************************************************************************* **
Constructor (NODE): Link this node into a list of nodes, and store the given data.
** ********************************************************************************************************* */
template<class MyType>
ListItem<MyType>::ListItem(MyType & data,ListItem<MyType> * pPrev,ListItem<MyType> * pNext)
{
	//memcpy((void *)&_data,(void *)&data,sizeof(MyType));
	_data = data;
	_pPrev = pPrev;
	_pNext = pNext;
}


#endif
