/* 
   resourceFileEditor.cc - resource file editor class implementation

   Will 18, 1999-2003

   This file is part of blackhome.

   blackhome is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   blackhome is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with blackhome; see the file COPYING.  If not, write to the Free
   Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ResourceFileEditor.h"
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>


#define debug


extern "C"
{
   void quitCallback ( Widget, XtPointer, XtPointer );
}


ResourceFileEditor::ResourceFileEditor ( char *homeDir, XtPointer app_resources, const char *name, Widget parent, XtAppContext app )
{
	Arg args [ 6 ];   
	int n;
	myResources = ( rfeResources * ) app_resources;
	_popup = parent;
	_app = app;
	rfeObjectData.redMin = 0;	// colour deviation from specified search value, lower bound
	rfeObjectData.redMax = 0;	// deviation, upper bound ( 0 <= value <= 1 )
	rfeObjectData.greenMin = 0;
	rfeObjectData.greenMax = 0;
	rfeObjectData.blueMin = 0;
	rfeObjectData.blueMax = 0;
	rfeObjectData.globalColourBound = 0;
	_searchXdefaults = 1;		// true
	_searchGtkdefaults = 1;
	_searchAppdefaults = 1;
	_searchMenus = 1;
	_searchStyles = 1;
	_returnSize = 8;			// hex rrggbb
	_fReturnSize = 4;			// float i.dddd
	_appendSearch = 0;			// force new search list build in 'find'
	_defaultColourType = 1;		// 'rgb:'
	_fallbackColourType = 1;	// 'rgb:'
	_execOnModify = 0;			// don't autoexec application when its resource file gets modified
	_shadeReplace = 0;			// do exact colour replacement, not relative intensity
	_shadeOffset = 0;			// do not force shade-scaled colour deviation from calculated intensity
	_generateHue = 0;			// just change shade on 'colour replace relative'
	_saveResourceMods = 0;		// don't keep 'undo' file
	_undoFileNum = 0;
	_useFirstResource = 0;		// first specified resource replaces all targets found
	_keepOldSearchTargets = 0;	// discard previous files to search in for each new 'find'
	_searchType = 0;			// default to colour; 1 = label
	strcpy ( _homeDir, homeDir );	

	if ( readPrefs ( _homeDir ) != 0 )
	{
		tablet ( "ResourceFileEditor_error", "Initialisation error",
 					".blackhome/config/resourceFileEditorrc nmissing:\nwriting default" );
		writePrefs ( _homeDir );
	}

	listApps ( _homeDir );
	listThemes ( _homeDir, myResources );
	listAliases ( _homeDir, myResources );

	if ( rfeObjectData.themes.Count() < myResources->maxMenuItems )
		_maxThemeMenuItems = rfeObjectData.themes.Count();
	else
		_maxThemeMenuItems = myResources->maxMenuItems;

	if ( rfeObjectData.apps.Count() < myResources->maxMenuItems )
		_maxAppMenuItems = rfeObjectData.apps.Count();
	else
		_maxAppMenuItems = myResources->maxMenuItems;

	_form = XtCreateManagedWidget ( "form" , formWidgetClass, _popup, NULL, 0 );

	n = 0;
	XtSetArg ( args[n], XtNlabel, ( char * ) currentTheme ( _homeDir,
				( void * ) myResources ) ); n++;

	XtSetArg ( args[n], XtNwidth, myResources->topLabelWidth ); n++;

	_label = XtCreateManagedWidget ( "label", labelWidgetClass, _form, args, n );

	// 'front-page' buttons
	
	_button = XtVaCreateManagedWidget ( "Quit", commandWidgetClass, _form,
										XtNfromVert, _label, NULL );

	XtAddCallback ( _button, XtNcallback, &ResourceFileEditor::quitCallback, 
					( XtPointer ) this );

	_button = XtVaCreateManagedWidget ( "Find", commandWidgetClass, _form,
										XtNfromVert, _label,
										XtNfromHoriz, _button, NULL );

	XtAddCallback ( _button, XtNcallback, &ResourceFileEditor::findCallback, 
					( XtPointer ) this );

	_button = XtVaCreateManagedWidget ( "Opts", commandWidgetClass, _form,
										XtNfromVert, _label,
										XtNfromHoriz, _button, NULL );

	XtAddCallback ( _button, XtNcallback, &ResourceFileEditor::optsCallback, 
					( XtPointer ) this );

	_button2 = XtVaCreateManagedWidget ( "View", commandWidgetClass, _form,
										XtNfromVert, _button, NULL );

	XtAddCallback ( _button2, XtNcallback, &ResourceFileEditor::viewCallback, 
					( XtPointer ) this );			

	_button2 = XtVaCreateManagedWidget ( "Seln", commandWidgetClass, _form,
										XtNfromVert, _button,
										XtNfromHoriz, _button2, NULL );

	XtAddCallback ( _button2, XtNcallback, &ResourceFileEditor::selnCallback, 
					( XtPointer ) this );

	_button2 = XtVaCreateManagedWidget ( "Exec", commandWidgetClass, _form,
										XtNfromVert, _button,
										XtNfromHoriz, _button2, NULL );

	XtAddCallback ( _button2, XtNcallback, &ResourceFileEditor::execCallback, 
					( XtPointer ) this );			
}


ResourceFileEditor::~ResourceFileEditor()
{
	cleanAvailable();	// remove available apps, themes and aliases lists
	cleanSelApps();		// remove list of selected apps to use as resource file search base
	cleanSelThemes();	// remove list of selected themes to use as resource file search base
	cleanSearchInFiles();	// remove list of selected target reosurce files
	cleanOperate();		// remove list of target resource files on which to operate
	cleanReplace();

	XtDestroyWidget ( _popup );
}


// Callback functions

void ResourceFileEditor::quitCallback ( Widget w, XtPointer clientData, XtPointer callData )
{
	ResourceFileEditor *resourceFileEditor = ( ResourceFileEditor * ) clientData;
	resourceFileEditor->quit ( w, clientData );
}    


void ResourceFileEditor::findCallback ( Widget w, XtPointer clientData, XtPointer )
{
	ResourceFileEditor * resourceFileEditor = ( ResourceFileEditor * ) clientData;
	resourceFileEditor->find ( w, clientData );
}


void ResourceFileEditor::optsCallback ( Widget w, XtPointer clientData, XtPointer )
{
	ResourceFileEditor * resourceFileEditor = ( ResourceFileEditor * ) clientData;
	resourceFileEditor->opts ( w, 0 );
}


void ResourceFileEditor::selnCallback ( Widget w, XtPointer clientData, XtPointer )
{
	ResourceFileEditor * resourceFileEditor = ( ResourceFileEditor * ) clientData;
	resourceFileEditor->seln ( w, 0 );
}


void ResourceFileEditor::viewCallback ( Widget w, XtPointer clientData, XtPointer )
{
	ResourceFileEditor * resourceFileEditor = ( ResourceFileEditor * ) clientData;
	resourceFileEditor->view ( w, 0 );
}


void ResourceFileEditor::execCallback ( Widget w, XtPointer clientData, XtPointer )
{
	ResourceFileEditor * resourceFileEditor = ( ResourceFileEditor * ) clientData;
	resourceFileEditor->exec ( w, clientData );
}


// class member functions for callback functions

void ResourceFileEditor::quit ( Widget, XtPointer clientData )
{
// XtDestroyWidget?
	exit (0);
}


int ResourceFileEditor::find ( Widget, XtPointer clientData )
{
	int selectedItemNum;
	int error = 0;
	char resource [ MAXCHARS ];
	char textString [ EDITBUFSIZE ];
	char * temp;
	float redLo, redHi, greenLo, greenHi, blueLo, blueHi;
	unsigned int aliasFlag;

	ColourSpecConvert *myConvert = new ColourSpecConvert;
	myConvert->setfReturnSize ( _fReturnSize );
	myConvert->setReturnSize ( _returnSize );
		
	error = myConvert->setRgbTxtPath ( myResources->rgbTxtDir );

	if ( error )
		reportColourError ( myResources->rgbTxtDir, error );

	redLo = redHi = greenLo = greenHi = blueLo = blueHi = 0;
	aliasFlag = 0;

	strcpy ( textString, _homeDir );
	strcat ( textString, "/" );
	strcat ( textString, myResources->resourceDir );

	error = myConvert->setResourceFilePath ( textString );

	if ( error )
	{
		strcat ( textString, "Couldn't access file\n'.colourAliasList'.\n");
		tablet ( "ResourceFileEditor_error", "Error", textString );
		delete myConvert;
		return 1;
	}

	if ( rfeObjectData.selThemes.Count() == 0 )
	{	// search app resource files for the current theme by default
		temp = new char [ ( strlen ( currentThemeName ) ) + 1 ];
		strcpy ( temp, currentThemeName );
		rfeObjectData.selThemes.Add ( temp );
	}

	ListItem < char * > *selThemeItem = rfeObjectData.selThemes.GetHead();
	ListItem < char * > *selAppItem = rfeObjectData.selApps.GetHead();

	if ( !_keepOldSearchTargets )
		cleanSearchInFiles();	// n.b existing selections to operate on are in opOn list

	while ( selThemeItem )
	{	// build target list of files to search
		if ( _searchXdefaults )
		{
			temp = new char [ ( MAXCHARS ) + 1 ];
			strcpy ( temp, myResources->XdefaultsDir );
			strcat ( temp, "/.Xdefaults-" );
			strcat ( temp, selThemeItem->Get() );
			rfeObjectData.targets.Add ( temp );	// .Xdefaults-themeName

			temp = new char [ ( MAXCHARS ) + 1 ];
			strcpy ( temp, myResources->XdefaultsDir );
			strcat ( temp, "/.Xresources-" );
			strcat ( temp, selThemeItem->Get() );
			rfeObjectData.targets.Add ( temp );	// .Xresources-themeName
		}

		if ( _searchGtkdefaults )
		{
			temp = new char [ ( MAXCHARS ) + 1 ];
			strcpy ( temp, myResources->XdefaultsDir );
			strcat ( temp, "/gtk/gtkrc-" );
			strcat ( temp, selThemeItem->Get() );
			rfeObjectData.targets.Add ( temp );	// themeName.menu
		}
				
		if ( _searchMenus )
		{
			temp = new char [ ( MAXCHARS ) + 1 ];
			strcpy ( temp, myResources->menusDir );
			strcat ( temp, "/" );
			strcat ( temp, selThemeItem->Get() );
			strcat ( temp, ".menu" );
			rfeObjectData.targets.Add ( temp );	// themeName.menu
		}
		
		if ( _searchStyles )
		{
			temp = new char [ ( MAXCHARS ) + 1 ];		
			strcpy ( temp, myResources->stylesDir );
			strcat ( temp, "/" );
			strcat ( temp, selThemeItem->Get() );
			rfeObjectData.targets.Add ( temp );	// style ( themeName )
		}
			
		if ( _searchAppdefaults )
		{
			while ( selAppItem )
			{
				temp = new char [ ( MAXCHARS ) + 1 ];
				strcpy ( temp, myResources->appResDir );
				strcat ( temp, "/" );
				strcat ( temp, selAppItem->Get() );
				strcat ( temp, "-" );
				strcat ( temp, selThemeItem->Get() );
				rfeObjectData.targets.Add ( temp );	// appResource-themeName
				selAppItem = selAppItem->Next();
			}
		
			selAppItem = rfeObjectData.selApps.GetHead();
			selThemeItem = selThemeItem->Next();
		}
		else
			selThemeItem = selThemeItem->Next();
	}	

	selectedItemNum = menu ( "resourceFileEditor", "Find resource","ResourceType",
								"Colour,Label,Exit,");

	if ( selectedItemNum < 3 ) // fall through on 'exit'; only need 1 popup box then
	{							
		if ( !_appendSearch )
			cleanOperate();		// remove previous operation targets
			
		tablet ( "Find_resource", "Enter a resource specification", "" );

		if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
		{
			currentEditLine[0] = 0;
			delete myConvert;
			return 1;
		}
		else
		{
			strcpy ( resource, currentEditLine );
			currentEditLine [ 0 ] = 0;
		}
	
		if ( strlen ( resource ) == 0 )
		{
			delete myConvert;
			return 1;
		}
	}

	switch ( selectedItemNum )
	{	// 'preprocess' the entered resource string, check for errors
		case 1:
		{	// colour - use colourSpecConvert to check format						
			aliasFlag = 0;	// set if using a colour alias
			_searchType = 0;	// search for colour resource
			
			if ( strlen ( resource ) < MAXCHARS )
		        error = myConvert->setColour ( resource );
			else
			{
				memError ( MAXCHARS, "find" );
				delete myConvert;
				return 1;
			}

			if ( error )
			{	// query for alias
				int oldError = error;
								
				ListItem < char * > *aliasItem = rfeObjectData.aliases.GetHead();
				error = 1;
				
				while ( aliasItem )
				{
					if( strcmp ( resource, aliasItem->Get() ) == 0 )

						error = 0;

					aliasItem = aliasItem->Next();
				}

				if ( error )
				{
					strcpy ( textString, resource );
					strcat ( textString, ".\nFailed check for valid X colour specification.\n" );
					strcat ( textString, "Failed check for valid blackhome colour alias.\n" );
					tablet ( "ResourceFileEditor_error", "Colour resource error", textString );
					
					reportColourError ( resource, oldError );					
					currentEditLine [ 0 ] = 0;
					delete myConvert;
					return 1;
				}

				aliasFlag = 1;
				currentEditLine [ 0 ] = 0;
			}
			else
			{ // modify for alias, don't have ints, do char match
				if ( !aliasFlag )
				{
					reportColour ( myConvert );	// continue with find

					if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
					{
						delete myConvert;
						return 1;
					}
				}
				else
				{	// colourAlias specifies a range, no variation of limits allowed
					redLo  = redHi = greenLo = greenHi = blueLo = blueHi = 0;				
				}
			}
		}

		break;
				
		case 2:
		{	// label
			aliasFlag = 0;	// since we didn't enter a colour
			_searchType = 1;	// default to search for colour by label for now
			// perhaps request type later
		}	// no input processing as yet

		break;
			
		default:
			delete myConvert;
			return 0;
	}

	// now find item in files
	char title [ strlen ( resource ) + 6 ];
	strcpy ( title, "Find " );
	strcat ( title, resource );

	tablet ( "ResourceFileEditor", title, "Press Ok and wait...\nor Cancel to quit find" );

	if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
	{
		currentEditLine [ 0 ] = 0;
		delete myConvert;
		return 1;
	}

	currentEditLine [ 0 ] = 0;

	FILE * resourceFile;
	int count;
	char fileName [ MAXCHARS ];
	char resourceLabel [ MAXCHARS ];
	char word [ MAXCHARS ];
	char partLine [ MAXCHARS ];
	char c;
	unsigned int numResources, targetNum, totalTargets;
	unsigned int ptr;
	int posInLine;	// start position of found item in string
	int length;
	bool isGnomeColour;

	numResources = 0;
	totalTargets = ( unsigned int ) rfeObjectData.targets.Count();
	
	// strip spaces to avoid a match failure
	ptr = 0;
	posInLine = 0;
	
	while ( resource [ ptr ] == ' ' )
		ptr++;

	while ( ptr < strlen ( resource ) )
	{
		word [ posInLine ] = resource [ ptr ];
		ptr++;
		posInLine++;
	}
	
	word [ posInLine ] = 0;

	ptr = strlen ( word ) - 1;
	
	while ( word [ ptr ] == ' ' )
		ptr--;
		
	for ( posInLine = 0; posInLine <= ( int ) ptr; posInLine++ )
		resource [ posInLine ] = word [ posInLine ];
		
	resource [ posInLine ] = 0;

	totalTargets = ( unsigned int ) rfeObjectData.targets.Count();
	
	for ( targetNum = 0; targetNum < totalTargets; targetNum++ )
	{
		strcpy ( fileName, _homeDir );
		strcat ( fileName, "/" );
		strcat ( fileName, rfeObjectData.targets[ targetNum ] );

		if ( ( resourceFile = fopen ( fileName, "r" ) ) == NULL )
		{	// for all files in targets: open file, check for resource
			strcpy ( textString, "File open failed: " );
			strcat ( textString, fileName );
			tablet ( "ResourceFileEditor_error", "Error", textString );
			delete myConvert;
			return 1;
		}
		else	
		{
			c = 0;		
			count = 0;

			while ( !feof ( resourceFile ) )		
			{
				fread ( &c, 1, 1, resourceFile );
				if ( count >= EDITBUFSIZE )
				{
					memError ( EDITBUFSIZE, "reading resource file" );
					delete myConvert;
					return 1;
				}
				else
					textString [ count ] = c;

				if ( c == 0x0a )	// got current line			
				{
					textString [ count + 1 ] = 0;

					if ( ( textString [ 0 ] == '#' ) || ( textString [ 0 ] == '!' ) )
						error = 1;	// commented line
					else
						error = 0;

					if ( !error )
					{
						// now we have to strip words: switch on case of 'find what'
						switch ( selectedItemNum )
						{
							case 1:
							{	// colour
								ptr = 0;
								posInLine = 0;
								length = strlen ( textString );
								word [ 0 ] = 0;

								// is it a gnome color format in a gtkrc file?
								if (strstr ( fileName, "gtkrc-" ) != NULL )
								{
									char gtkresource [ MAXCHARS ];
									int pos, ptr2, length;
									float gtkR,gtkG,gtkB;

									pos = strstr ( textString, "] = {" ) - textString + 4;	// start of colour spec

									if ( pos < strlen ( textString ) )
									{
										length = strlen ( textString ) - pos - 1;
										char word [ length ];

										for ( ptr2 = 0; ptr2 < length; ptr2++ )
											word [ ptr2 ] = textString [ ptr2 + pos ];
	
										word [ length ] = 0;
										sscanf ( word, "{ %f, %f, %f }", &gtkR, &gtkG, &gtkB );
										snprintf ( gtkresource, sizeof ( gtkresource ), "rgbi:%f/%f/%f", gtkR, gtkG, gtkB );

										if ( myConvert->setColour ( gtkresource ) == 0 )	// or it doesn't know about it...
											snprintf ( word, sizeof ( gtkresource ), myConvert->rgbiColour());
										else
										{
											snprintf ( textString, sizeof ( textString ), "GTK colour format unrecognised\nshould be red, green, blue\n\
floating point range 0-1\nfound GTK colour %s", word );
											tablet ( "ResourceFileEditor_error", "Error", textString );
											delete myConvert;
											return 1;
										}

										isGnomeColour = TRUE;

										for (ptr2 = 0; ptr2 <= pos; ptr2++)
											resourceLabel [ ptr2 ] = textString [ ptr2 ];

										resourceLabel [ ptr2 ] = 0;
										// add resource if match to user input resource succeeds
										numResources += addColourTarget( ( char * ) resource, myConvert, aliasFlag,
											( char * ) fileName, ( char * ) resourceLabel, ( char * ) word,
											rfeObjectData.redMin,rfeObjectData.redMax,rfeObjectData.greenMin,
											rfeObjectData.greenMax,rfeObjectData.blueMin, rfeObjectData.blueMax );
									}
								}
								else
									isGnomeColour = FALSE;

								if ( isGnomeColour == FALSE )	// might be a gtkrc with #colour definition
								{
									while ( posInLine < length )
									{
										c = textString [ posInLine ];
										partLine [ posInLine ] = c;
										// ugh. need to do this a la case 'label'...
										if ( ( ( c >= 'A' ) && ( c <= 'Z' ) ) || ( 
										( c >= 'a' ) && ( c <= 'z' ) ) || ( c == '#' )
										|| ( c == '/' ) || ( c == '.' ) || ( c == '*' )
										|| ( c == ' ' ) || ( c == '-' ) )	// s/b strstr
										{	// commas and underscores irrelevant to colour spec, and _ignore_ colons till later. Yes, horrible.
											// pass hyphens since they're argument flags
											word [ ptr ] = c;
											ptr++;
										}
										else
										{
											if ( ( c >= '0' ) && ( c <= '9' ) || ( c == ':' ) )
											{	// valid 'extension'
												word [ ptr ] = c;
												ptr++;
											}
										}
										posInLine++;

										if ( ( c == ' ' ) || ( c == '\n' ) || ( c == '\t' ) )
										{	// detect space, return. tab
											word [ ptr ] = 0;

											if ( strlen ( word ) != 0 )
											{
												if ( word [ strlen ( word ) - 1 ] == ' ' )
													word [ strlen ( word ) - 1 ] = 0;	// cut trailing space if any

												error = myConvert->setColour ( word );

												if ( !error )
												{	// found valid std colour statement
													partLine [ posInLine ] = 0;

													if ( strstr ( fileName, ".menu" ) != NULL )	// see if this is a blackbox menu
													{	// resourceLabel invalid if this is a .menu type file
														// remove extra occurrence of colour name at end of partLine
														unsigned int linePtr, tPtr;

														tPtr = 0;
														linePtr = (unsigned int ) ( strstr ( partLine, word ) - partLine );

														while ( tPtr < linePtr )
														{
															resourceLabel [ tPtr ] = partLine [ tPtr ];
															tPtr++;
														}

														resourceLabel [ tPtr ] = 0;													
													}
													// add resource if match to user input resource succeeds
													numResources += addColourTarget( ( char * ) resource, myConvert, aliasFlag,
																( char * ) fileName, ( char * ) resourceLabel, ( char * ) word,
																rfeObjectData.redMin,rfeObjectData.redMax,rfeObjectData.greenMin,
																rfeObjectData.greenMax,rfeObjectData.blueMin, rfeObjectData.blueMax );
												}
												else
												{	// successful colour spec means preceding word is resource using this colour
													if ( strcmp ( word, " " ) != 0 )
														strcpy ( resourceLabel, word );	// if the next word succeeds
												}

												ptr = 0;
											}

											word [ 0 ] = 0;
										}
									}
								}	// end standard colour format
							}

							break;

							case 2:
							{	// label
								char partLine1 [ MAXCHARS ];
								char partLine2 [ MAXCHARS ];
								char temp [ MAXCHARS ];
								char altResource [ MAXCHARS ];								

								// if this is a gtk resource we skip extraction
								if (strstr ( fileName, "gtkrc-" ) == NULL )	// not a gtk resource
								{
									posInLine = 0;
									ptr = 0;
									length = strlen ( textString );
									resourceLabel [ 0 ] = 0;
									error = 1;

									strcpy ( altResource, resource );

									if ( ( ( int ) resource [ 0 ] > 64 ) || ( (int ) resource [ 0 ] < 91 ) )
										altResource [ 0 ] += 32;
									else
										altResource [ 0 ] -= 32;	// presume it was lowercase
								}	// else gnome resource (extracted by label), continue

								int count = 0;
								long diff;
								int appNameType, foundType;	// switch resource / altResource

								error = 1;
								appNameType = 1;	// default to supplied resource name
								foundType = 0;

								if ( strstr ( textString, resource ) != NULL )
								{
									error = 0;
									appNameType = 1;
								}
								else
								{
									if ( strstr ( textString, altResource ) != NULL )
									{
										error = 0;
										appNameType = 2;
									}
								}

								if ( !error )
								{
									// is it a gnome color?
									if (strstr ( fileName, "gtkrc-" ) != NULL )
									{
										char gtkresource [ MAXCHARS ];
										int pos, ptr2, length;
										float gtkR,gtkG,gtkB;

										pos = strstr ( textString, "] = {" ) - textString + 4;	// start of colour spec

										if ( pos < strlen ( textString ) )
										{
											length = strlen ( textString ) - pos - 1;
											char word [ length ];

											for ( ptr2 = 0; ptr2 < length; ptr2++ )
												word [ ptr2 ] = textString [ ptr2 + pos ];
	
											word [ length ] = 0;

											sscanf ( word, "{ %f, %f, %f }", &gtkR, &gtkG, &gtkB );
											snprintf ( gtkresource, sizeof ( gtkresource ), "rgbi:%f/%f/%f", gtkR, gtkG, gtkB );

											if ( myConvert->setColour ( gtkresource ) == 0 )	// or it doesn't know about it...
												snprintf ( word, sizeof ( gtkresource ), myConvert->rgbiColour());
											else
											{
												snprintf ( textString, sizeof ( textString ), "Gnome colour format unrecognised\nshould be red, green, blue\n\
floating point range 0-1\nfound Gnome colour equivalent:\n'%s'", word );
												tablet ( "ResourceFileEditor_error", "Error", textString );
												delete myConvert;
												return 1;
											}

											for (ptr2 = 0; ptr2 <= pos; ptr2++)
												resourceLabel [ ptr2 ] = textString [ ptr2 ];

											resourceLabel [ ptr2 ] = 0;
											// word is rgbi gnome colour, check resourceLabel for resource
											if ( strstr ( resourceLabel, resource ) != NULL )
												numResources += addColourTarget( ( char * ) word, myConvert, aliasFlag,
													( char * ) fileName, ( char * ) resourceLabel, ( char * ) word,
													rfeObjectData.redMin,rfeObjectData.redMax,rfeObjectData.greenMin,
													rfeObjectData.greenMax,rfeObjectData.blueMin, rfeObjectData.blueMax );
										}
									}
									else	// all std format colours
									{
										if ( appNameType == 1 )
											diff  = strstr ( textString, resource ) - textString;
										else
											diff  = strstr ( textString, altResource ) - textString;

										while ( count < diff )
										{
											partLine1 [ count ] = textString [ count ];
											count++;
										}

										partLine1 [ count ] = 0;

										strcpy ( resourceLabel, partLine1 );

										if ( appNameType == 1 )
											strcat ( resourceLabel, resource );
										else
											strcat ( resourceLabel, altResource );

										count += strlen ( resource );	// same as altR
										ptr = 0;

										while ( count < ( int ) strlen ( textString ) )
										{
											temp [ ptr ] = textString [ count ];
											count++;
											ptr++;
										}

										temp [ ptr ] = 0;
										count = 0;

										if ( strstr ( temp, resource ) != NULL )
										{	// another mention of the resource
											foundType = 1;
											diff  = strstr ( temp, resource ) - temp;

											while ( count < diff )
											{
												partLine2 [ count ] = temp [ count ];
												count++;
											}

											partLine2 [ count ] = 0;
										}
										else
										{
											if ( strstr ( temp, altResource ) != NULL )
											{	// another mention of the resource
												foundType = 2;
												diff  = strstr ( temp, altResource ) - temp;

												while ( count < diff )
												{
													partLine2 [ count ] = temp [ count ];
													count++;
												}

												partLine2 [ count ] = 0;
											}
											else
												partLine2 [ 0 ] = 0;
										}

										// found mention of resource in remainder of line from file
										if ( strlen ( partLine2 ) != 0 )
										{
											strcat ( resourceLabel, partLine2 );

											if ( foundType == 1 )
												strcat ( resourceLabel, resource );
											else
												strcat ( resourceLabel, altResource );
										}

										ptr = strlen ( resourceLabel ); // might as well start here
										count = 0;

										while ( ptr < strlen ( textString ) )
										{
											temp [ count ] = textString [ ptr ];
											ptr++;
											count++;
										}

										temp [ count ] = 0;
										ptr = 0;
										posInLine = 0;

										while ( ptr < strlen ( temp ) )
										{
											c = temp [ ptr ];
											word [ posInLine ] = c;

											if ( ( c == ' ' ) || ( c == '\t' ) || ( c == '\n' )
													|| ( c == '}' ) ) // word terminators
											{
												word [ posInLine ] = 0;

												if ( _searchType == 1 ) // colour search
												{
													error = myConvert->setColour ( word );

													if ( !error )
													{	// add instance of found colour resource
														// copy base resource label to new string
														char newLabel [ MAXCHARS ];
														strcpy ( newLabel, resourceLabel );												
														diff  = strstr ( temp, word ) - temp;
														count = 0;

														while ( count < diff )
														{	// get remainder of label before colour
															partLine1 [ count ] = temp [ count ];
															count++;
														}

														partLine1 [ count ] = 0;
														strcat ( newLabel, partLine2 );
														strcat ( newLabel, partLine1 );

														if ( aliasFlag )
														{	// colourAlias specifies a range, no variation of limits allowed
															redLo  = redHi = greenLo = greenHi = blueLo = blueHi = 0;
														}

														numResources += addColourTarget( ( char * ) word, myConvert, aliasFlag,
																	( char * ) fileName, ( char * ) newLabel, ( char * ) word,
																rfeObjectData.redMin,rfeObjectData.redMax,rfeObjectData.greenMin,
																rfeObjectData.greenMax, rfeObjectData.blueMin, rfeObjectData.blueMax );

													} // end was 'word' colour resource?
												}	// end colour search by label

												// other types should be sorted here
												// 'addOtherTarget', depending...

												word [ 0 ] = 0;
												posInLine = -1;
											}	// end found word in line

											ptr++;
											posInLine++;
										} // end while ptr < strlen(temp)

									}

									resourceLabel [ 0 ] = 0;
								}	// end if resource occurs in line
							}	// end case 2: find 'label'
							
							break;
							
							default:
							break;
						}	// end switch 'colour,label'
					}	// end if line !comment or #include

					count = -1;	// reset for new line
					textString [ 0 ] = 0;
				}	// end got current line

				count++;
			}	// end of file, done
			
			fclose ( resourceFile );
		}	// end check file available
	}	// done all target files
	
	// warn on no occurrences found
	if ( numResources == 0 )
	{
		char title [ strlen ( resource ) + 6 ];
		strcpy ( title, "Find " );
		strcat ( title, resource );
		char msg [ strlen ( resource ) + 26 ];
		strcpy ( msg, "No occurrences of " );
		strcat ( msg, resource );
		strcat ( msg, " found." );
		tablet ( "ResourceFileEditor", title, msg );
	}
	
	delete myConvert;
	return 0;	
}


int ResourceFileEditor::opts ( Widget, XtPointer clientData )
{
	char textString [ EDITBUFSIZE ];
	char numString [ MAXCHARS ];
	int selectedItemNum;
	
	strcpy ( textString, "Colour,Resource,Search,Undo,Replace,Save,Exit" );	
	selectedItemNum = menu ( "resourceFileEditor", "Options","Preferences",
								textString );
									
	switch ( selectedItemNum )
	{
		case 1:
		{	// colour options
			int colourItemNum;

			colourItemNum = menu ( "colourSpecConvert", "Colour", "Options",
								"Depth,Digits,Deviation,Exit," );

			switch ( colourItemNum )
			{
				case 1:
				{	// depth
					int depthItemNum;

					strcpy ( textString, "Current colour depth is " );
					sprintf ( numString, "%u", _returnSize );
					strcat ( textString, numString );
					strcat ( textString, " bits per pixel\n" );
					strcat ( textString, "Change this value?" );
					tablet ( "ColourSpecConvert", "colour depth", textString );
					textString [ 0 ] = 0;
		
					if ( strcmp ( currentEditLine, "tabletCancel" ) != 0 )
					{
						depthItemNum = menu ( "colourSpecConvert", "Colour depth",
												"bits/pixel", "4,8,12,16," );
			
						_returnSize = 4 * depthItemNum;
					}
				}

				break;

				case 2:
				{	// f.p. digits
					int floatItemNum;

					strcpy ( textString, "Current significant digits: " );
					sprintf ( numString, "%u", _fReturnSize );
					strcat ( textString, numString );
					tablet ( "ColourSpecConvert", "significant digits", textString );
					textString [ 0 ] = 0;

					if ( strcmp ( currentEditLine, "tabletCancel" ) != 0 )
					{
						floatItemNum = menu ( "colourSpecConvert", "fp digits",
												"number", "0,1,2,3,4,5,6,7," );
								
						_fReturnSize = floatItemNum - 1;
					}
				}
				
				break;
				
				case 3:
				{	// allowed deviation from specified colour for search
					float t1, t2, t3, t4, t5, t6, t7;
					float error;

					strcpy ( textString, "red-   " );
					sprintf ( numString, "%f\n", rfeObjectData.redMin );
					strcat ( textString, numString );
					strcat ( textString, "red+   " );
					sprintf ( numString, "%f\n", rfeObjectData.redMax );
					strcat ( textString, numString );
					strcat ( textString, "green- " );
					sprintf ( numString, "%f\n", rfeObjectData.greenMin );
					strcat ( textString, numString );
					strcat ( textString, "green+ " );
					sprintf ( numString, "%f\n", rfeObjectData.greenMax );
					strcat ( textString, numString );
					strcat ( textString, "blue-  " );
					sprintf ( numString, "%f\n", rfeObjectData.blueMin );
					strcat ( textString, numString );
					strcat ( textString, "blue+  " );
					sprintf ( numString, "%f\n", rfeObjectData.blueMax );
					strcat ( textString, numString );
					strcat ( textString, "global " );
					sprintf ( numString, "%f\n", rfeObjectData.globalColourBound );
					strcat ( textString, numString );
					
					strcpy ( textString, tablet ( "ColourSpecConvert",
													"r,g,b deviations", textString ) );

					if ( strcmp ( currentEditLine, "tabletCancel" ) != 0 )
					{
						sscanf ( textString, "%*s %f %*s %f %*s %f %*s %f %*s %f %*s %f %*s %f",
									&t1, &t2, &t3, &t4, &t5, &t6, &t7 );

						numString [ 0 ] = 0;
						textString [ 0 ] = 0;
						error = 0;

						if ( t7 < 0 )
							t7 = 0;

						if ( t7 > 1 )
							t7 = 1;

						rfeObjectData.globalColourBound = t7;

						if ( t7 != 0 )
							t1 = t2 = t3 = t4 = t5 = t6 = t7;

						if ( ( t1 < 0 ) || ( t1 > 1 ) )
						{
							error = t1;
							strcat ( textString, "red- error: " );
							sprintf ( numString, "%f\n", error );
							strcat ( textString, numString );
						}
						else
							rfeObjectData.redMin = t1;					
					
						if ( ( t2 < 0 ) || ( t2 > 1 ) )
						{
							error = t2;
							strcat ( textString, "red+ error: " );
							sprintf ( numString, "%f\n", error );
							strcat ( textString, numString );
						}
						else
							rfeObjectData.redMax = t2;
						
						if ( ( t3 < 0 ) || ( t3 > 1 ) )
						{
							error = t3;
							strcat ( textString, "green- error: " );
							sprintf ( numString, "%f\n", error );
							strcat ( textString, numString );
						}
						else
							rfeObjectData.greenMin = t3;
											
						if ( ( t4 < 0 ) || ( t4 > 1 ) )
						{
							error = t4;
							strcat ( textString, "green+ error: " );
							sprintf ( numString, "%f\n", error );
							strcat ( textString, numString );
						}
						else
							rfeObjectData.greenMax = t4;
					
						if ( ( t5 < 0 ) || ( t5 > 1 ) )
						{
							error = t5;
							strcat ( textString, "blue- error: " );
							sprintf ( numString, "%f\n", error );
							strcat ( textString, numString );
						}
						else
							rfeObjectData.blueMin = t5;
					
						if ( ( t6 < 0 ) || ( t6 > 1 ) )
						{
							error = t6;
							strcat ( textString, "blue+ error: " );
							sprintf ( numString, "%f\n", error );
							strcat ( textString, numString );
						}
						else
							rfeObjectData.blueMax = t6;
					
						if ( error )
						{
							strcat ( textString, "Allowed deviation range exceeded:\n" );
							strcat ( textString, "0.0 <= deviation <= 1.0\n" );
							strcat ( textString, "Re-enter deviation values\n" );
							tablet ( "ResourceFileEditor_error", "Error", textString );
							textString [ 0 ] = 0;	
						}
					}
				}
				
				break;
				
			default:
				break;

			}	// end switch colourItemNum
		}
			
		break;

		case 2:
			{	// resource preferences - search in...
				int resourceItemNum;

				resourceItemNum = menu ( "resourceFileEditor", "Resources", "Options",
										 "Xdefaults,Gtk defaults,App defaults,Menus,Styles,Exit," );

				strcpy ( textString, "Current status: " );
								
				switch ( resourceItemNum )
				{
					case 1:
						{	// .Xdefaults and .Xresources
							if ( _searchXdefaults )
								strcat ( textString, " search ");
							else
								strcat ( textString, " don't search " );
							
							strcat ( textString, " .Xdefaults, .Xresources.\n" );
							strcat ( textString, "Change this option?\n" );
							
							tablet ( "ResourceFileEditor", "file search", textString );
							
							if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
							{
								currentEditLine[0] = 0;
								return 1;
							}
							else
							{
								_searchXdefaults = !_searchXdefaults;
								currentEditLine [ 0 ] = 0;
							}
						}
					
						break;
					
					case 2:
						{	// gtk defaults
							if ( _searchGtkdefaults )
								strcat ( textString, " search " );
							else
								strcat ( textString, " don't search " );
														
							strcat ( textString, "gtk defaults.\n" );
							strcat ( textString, "Change this option?\n" );
							
							tablet ( "ResourceFileEditor", "gtkrc search", textString );
							
							if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
							{
								currentEditLine[0] = 0;
								return 1;
							}
							else
							{
								_searchGtkdefaults = !_searchGtkdefaults;
								currentEditLine [ 0 ] = 0;
							}
						}
					
						break;
					
					case 3:
						{	// app defaults
							if ( _searchAppdefaults )
								strcat ( textString, " search\n");
							else
								strcat ( textString, " don't search\n" );
														
							strcat ( textString, "application defaults.\n" );
							strcat ( textString, "Change this option?\n" );
							
							tablet ( "ResourceFileEditor", "application search", textString );
							
							if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
							{
								currentEditLine[0] = 0;
								return 1;
							}
							else
							{
								_searchAppdefaults = !_searchAppdefaults;
								currentEditLine [ 0 ] = 0;
							}
						}
					
						break;

					case 4:
						{	// menus
							if ( _searchMenus )
								strcat ( textString, " search ");
							else
								strcat ( textString, " don't search " );
								
							
							strcat ( textString, "menus.\n" );
							strcat ( textString, "Change this option?\n" );
							
							tablet ( "ResourceFileEditor", "menu search", textString );
							
							if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
							{
								currentEditLine[0] = 0;
								return 1;
							}
							else
							{
								_searchMenus = !_searchMenus;
								currentEditLine [ 0 ] = 0;
							}
						}	
					
						break;
					
					case 5:
						{	// styles
							if ( _searchStyles )
								strcat ( textString, " search ");
							else
								strcat ( textString, " don't search " );
								
							strcat ( textString, "styles.\n" );
							strcat ( textString, "Change this option?\n" );
							
							tablet ( "ResourceFileEditor", "style search", textString );
							
							if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
							{
								currentEditLine[0] = 0;
								return 1;
							}
							else
							{
								_searchStyles = !_searchStyles;
								currentEditLine [ 0 ] = 0;
							}
						}
					
						break;
						
					default:
						break;
				}
			}
			
			break;
			
		case 3:
			{	// search preferences
				int searchItemNum;

				searchItemNum = menu ( "resourceFileEditor", "Search", "Options",
										 "Append search,Preserve targets,Exit," );

				strcpy ( textString, "Current status: " );
								
				switch ( searchItemNum )
				{
					case 1:
						{	// append or create new search
			
							if ( _appendSearch )
							{
								strcat ( textString, " append results of new search\n" );
								strcat ( textString, "to previous search results.\n");
							}
							else
							{
								strcat ( textString, " discard previous search results\n" );
								strcat ( textString, "and start afresh.\n" );
							}

							strcat ( textString, "Change this option?\n" );

							tablet ( "ResourceFileEditor", "append search", textString );

							if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
							{
								currentEditLine[0] = 0;
								return 1;
							}
							else
							{
								_appendSearch = !_appendSearch;
								currentEditLine [ 0 ] = 0;
							}
						}
						
						break;
						
					case 2:
						{	// preserve or discard previous target file list
							if ( _keepOldSearchTargets )
								strcat ( textString, "\nsearch in previously specified files.\n" );
							else
							{
								strcat ( textString, "\nuse blank search target list for\n" );
								strcat ( textString, "each new 'find' operation\n" );
							}

							strcat ( textString, "Change this option?\n" );

							tablet ( "ResourceFileEditor", "search target list", textString );

							if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
							{
								currentEditLine[0] = 0;
								return 1;
							}
							else
							{
								_keepOldSearchTargets = !_keepOldSearchTargets;
								currentEditLine [ 0 ] = 0;
							}
						}
						
						break;
				}	// end switch searchitemNum
			}	// end case 3: search preferences
			
			break;

		case 4:
			{	// save resource modifications
				strcpy ( textString, "Current status: " );
			
				if ( _saveResourceMods )
				{
					strcat ( textString, " save undo files in \n" );
					strcat ( textString, _homeDir );
					strcat ( textString, "/" );
					strcat ( textString, myResources->resourceDir );
					strcat ( textString, "\n" );
				}
				else
					strcat ( textString, " don't save undo files.\n" );

				strcat ( textString, "Change this option?\n" );

				tablet ( "ResourceFileEditor", "Options/undo", textString );

				if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
				{
					currentEditLine[0] = 0;
					return 1;
				}
				else
				{
					_saveResourceMods = !_saveResourceMods;
					currentEditLine [ 0 ] = 0;
				}
			}
			
			break;

		case 5:
			{	// replace options
				int replaceItemNum;
				
				replaceItemNum = menu ( "resourceFileEditor", "Replace",
											"type", "Autoexec,Colour,Exit," );
											
				switch ( replaceItemNum )
				{
					case 1:
						{
							strcpy ( textString, "Current status:\n" );
							strcat ( textString, "Execute on modify: " );
			
							if ( _execOnModify )
								strcat ( textString, "true\n" );
							else
								strcat ( textString, "false\n" );

							strcat ( textString, "Change this option?\n" );

							tablet ( "ResourceFileEditor", "autoexec", textString );

							if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
							{
								currentEditLine[0] = 0;
								return 1;
							}
							else
							{
								_execOnModify = !_execOnModify;
								currentEditLine [ 0 ] = 0;
							}
						}
						
						break;

					case 2:
						{
							int typeItemNum;
							char labelText [ MAXCHARS ];
							
							strcpy ( labelText, "Replacement type,Colour type,Force shade," );
							
							if ( _useFirstResource )
								strcat ( labelText, "Global replace," );
							else
								strcat ( labelText, "Local replace," );
								
							strcat ( labelText, "Derivation mode,Exit," );
							
							typeItemNum = menu ( "resourceFileEditor", "Colour", "Type", labelText );

							switch ( typeItemNum )
							{
	
								case 1:
									{
										strcpy ( textString, "Current replacement type:\n" );
										
										if ( _shadeReplace )
											strcat ( textString, "relative intensity\n" );
										else
											strcat ( textString, "exact specification\n" );
											
										strcat ( textString, "Change this option?\n" );

										tablet ( "ResourceFileEditor", "colour replacement type", textString );

										if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
										{
											currentEditLine[0] = 0;
											return 1;
										}
										else
										{
											_shadeReplace = !_shadeReplace;
											currentEditLine [ 0 ] = 0;
										}
									}
									
									break;
									
								case 2:
									{
										strcpy ( textString, "Current colour type: " );
							
										switch ( _defaultColourType )
										{
											case 0:
												strcat ( textString, "#" );
												break;
									
											case 1:
												strcat ( textString, "rgb" );
												break;
									
											case 2:
												strcat ( textString, "rgbi" );
												break;
									
											case 3:
												strcat ( textString, "named" );
										}
									
										tablet ( "ColourSpecConvert", "colour type", textString );
										textString [ 0 ] = 0;

										if ( strcmp ( currentEditLine, "tabletCancel" ) != 0 )
										{
											int defColItem, fallbackColItem;
							
											 defColItem = menu ( "ResourceFileEditor", "Colour default",
																	"type", "#,rgb,rgbi,named," );
						
											_defaultColourType = defColItem - 1;

											if ( _defaultColourType == 3 )
											{
												 fallbackColItem = menu ( "ResourceFileEditor", "Colour fallback",
																			"type", "#,rgb,rgbi," );
													
												_fallbackColourType = fallbackColItem - 1;
											}
											else
												_fallbackColourType = 1;
										}
										else
											return 1;
									}
									
									break;

								case 3:
									{
										char numString [ MAXCHARS ];
										
										strcpy ( textString, "Replace shade offset " );
										sprintf( numString, "%f", _shadeOffset );
										strcat ( textString, numString );
										strcat ( textString, " with..." );
										tablet ( "ResourceFileEditor", textString, "" );

										if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
										{
											currentEditLine [ 0 ] = 0;
											return 1;
										}
										else
										{
											sscanf ( currentEditLine, "%f", & _shadeOffset );
											currentEditLine [ 0 ] = 0;
											
											if ( _shadeOffset > 1 )
												_shadeOffset = 1;
												
											if ( _shadeOffset < -1 )
												_shadeOffset = -1;
												
											if ( !_shadeReplace )
											{
												strcpy ( textString, "Warning: shade replacement option\n" );
												strcat ( textString, "not selected: shade offset will\n" );
												strcat ( textString, "have no effect," );
												tablet ( "ResourceFileEditor","shade option error", textString );
												currentEditLine [ 0 ] = 0;	// no return
											}
										}
									}
									
									break;

								case 4:
									{
										strcpy ( textString, "Current replacement mode:\n" );
										
										if ( _useFirstResource )
										{
											strcat ( textString, "First specified replacement resource\n" );
											strcat ( textString, "substitutes for all target resources.\n" );
										}
										else
										{
											strcat ( textString, "Prompt for a replacement resource\n" );
											strcat ( textString, "when a new target resource is found.\n" );
										}
											
										strcat ( textString, "Change this option?\n" );

										tablet ( "ResourceFileEditor", "colour replacement type", textString );

										if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
										{
											currentEditLine[0] = 0;
											return 1;
										}
										else
										{
											_useFirstResource = !_useFirstResource;
											currentEditLine [ 0 ] = 0;
										}
									}
									
									break;
								
								case 5:
									{
										strcpy ( textString, "Colour derivation mode:\n" );
										
										if ( _generateHue )
											strcat ( textString, "Generate new colour.\n" );
										else
											strcat ( textString, "Shade replacement colour.\n" );
											
										strcat ( textString, "Change this option?\n" );

										tablet ( "ResourceFileEditor", "colour derivation mode", textString );

										if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
										{
											currentEditLine[0] = 0;
											return 1;
										}
										else
										{
											_generateHue = !_generateHue;
											currentEditLine [ 0 ] = 0;
										}
									}
									
									break;
								
								default:
									break;

							}	// end switch typeItemNum
						}	// end case 2: 'exec/replace/colour'
												
						break;
						
						default:
							break;
				}	// end switch replaceItemNum
			}	// end case 5: replace options
				
			break;

		case 6:
			{	// save preferences
				writePrefs( _homeDir );
			}
		
			break;
			
		default:
			break;	// Exit
	}
	
	return 0;
}


int ResourceFileEditor::seln ( Widget, XtPointer clientData )
{
	int count;
	int selectedItemNum;

	selectedItemNum = menu ( "resourceFileEditor_select", "Target","Target type",
								"Applications,Themes,Exit,");
	switch ( selectedItemNum )
	{
		case 1:
			{	// applications
				int appItemNum;
				unsigned int length;
				int menuStart, menuEnd, numMenuItems;
				
				menuStart = 0;
				count = 0;

				if ( _maxAppMenuItems < rfeObjectData.apps.Count() )
					numMenuItems = _maxAppMenuItems;
				else
					numMenuItems = rfeObjectData.apps.Count();

				menuEnd = numMenuItems;

				int keepLooping = 1;
			
				while ( keepLooping )
				{
					length = 0;
					
					for ( count = menuStart; count < numMenuItems; count++ )
					{
						length += strlen ( rfeObjectData.apps[ count ].name );
						length += 1; // account for ','
					}
				
					length += 15;	// and for 'All apps,' and 'Exit.\0'
				
					if ( length >= EDITBUFSIZE )
					{
						memError ( EDITBUFSIZE, "seln/apps" );
						return 1;					
					}
				
					strcpy ( currentEditLine, "All apps," );

					for ( count = menuStart; count < menuEnd; count++ )
					{			
						strcat ( currentEditLine, rfeObjectData.apps[ count ].name );
						strcat ( currentEditLine, "," );
					}

					if ( menuEnd < rfeObjectData.apps.Count() )
						strcat ( currentEditLine, "More," );
					else
						strcat ( currentEditLine, "Exit," );

					appItemNum = menu ( "resourceFileEditor", "Select Target",
										"Applications", currentEditLine );

					currentEditLine [ 0 ] = 0;

					if ( appItemNum < ( numMenuItems + 2 ) )
					{
						appItemNum += menuStart;
						keepLooping = 0;
					}
					else
					{
						if ( ( appItemNum + menuStart ) == ( rfeObjectData.apps.Count() + 2 ) )
						{
							keepLooping = 0;
							appItemNum = rfeObjectData.apps.Count() + 2;
						}
				
						menuStart += numMenuItems;
						menuEnd += numMenuItems;
					
						if ( menuEnd > rfeObjectData.apps.Count() )
						{
							menuStart = rfeObjectData.apps.Count() - numMenuItems;
							menuEnd = rfeObjectData.apps.Count();
						}
						
						if ( menuStart > ( rfeObjectData.apps.Count() - numMenuItems ) )
						{
							menuStart = 0;
							menuEnd = numMenuItems;
						}
					}
				}	

				if ( appItemNum == 1 )
				{	// add all app resource files to the target list
					for ( count = 0; count < rfeObjectData.apps.Count(); count++ )
						rfeObjectData.selApps.Add ( rfeObjectData.apps[ count ].resourceName );
				}
				else
				{	// add selected theme resource file to the target list
					if ( appItemNum < ( rfeObjectData.apps.Count() + 2 ) ) // skip 'Exit' - isn't a theme
						rfeObjectData.selApps.Add ( rfeObjectData.apps[ appItemNum - 2 ].resourceName );
				}
			}
			
			break;

		case 2:
			{	// themes
				int themeItemNum;
				unsigned int length;
				int menuStart, menuEnd, numMenuItems;

				menuStart = 0;
				count = 0;

				if ( _maxThemeMenuItems < rfeObjectData.apps.Count() )
					numMenuItems = _maxThemeMenuItems;
				else
					numMenuItems = rfeObjectData.themes.Count();

				menuEnd = numMenuItems;				

				int keepLooping = 1;
			
				while ( keepLooping )
				{
					length = 0;
					
					for ( count = menuStart; count < numMenuItems; count++ )
					{
						length += strlen ( rfeObjectData.themes[ count ] );
						length += 1; // account for ','
					}
				
					length += 17;	// and for 'All themes,' and 'Exit,' or 'more,', then '\0'
				
					if ( length >= EDITBUFSIZE )
					{
						memError ( EDITBUFSIZE, "seln/themes" );
						return 1;					
					}
				
					strcpy ( currentEditLine, "All themes," );

					for ( count = menuStart; count < menuEnd; count++ )
					{			
						strcat ( currentEditLine, rfeObjectData.themes[ count ] );
						strcat ( currentEditLine, "," );
					}

					if ( menuEnd < rfeObjectData.themes.Count() )
						strcat ( currentEditLine, "More," );
					else
						strcat ( currentEditLine, "Exit," );

					themeItemNum = menu ( "resourceFileEditor", "Select Target",
											"Themes", currentEditLine );

					currentEditLine [ 0 ] = 0;

					if ( themeItemNum < ( numMenuItems + 2 ) )
					{
						themeItemNum += menuStart;
						keepLooping = 0;
					}
					else
					{
						if ( ( themeItemNum + menuStart ) == ( rfeObjectData.themes.Count() + 2 ) )
						{
							keepLooping = 0;
							themeItemNum = rfeObjectData.themes.Count() + 2;
						}
				
						menuStart += numMenuItems;
						menuEnd += numMenuItems;
					
						if ( menuEnd > rfeObjectData.themes.Count() )
						{
							menuStart = rfeObjectData.themes.Count() - numMenuItems;
							menuEnd = rfeObjectData.themes.Count();
						}
						
						if ( menuStart > ( rfeObjectData.themes.Count() - numMenuItems ) )
						{
							menuStart = 0;
							menuEnd = numMenuItems;
						}
					}
				}	

				if ( themeItemNum == 1 )
				{	// add all theme resource files to the target list
					for ( count = 0; count < rfeObjectData.themes.Count(); count++ )
						rfeObjectData.selThemes.Add ( rfeObjectData.themes[ count ] );
				}
				else
				{	// add selected theme resource file to the target list
					if ( themeItemNum < ( rfeObjectData.themes.Count() + 2 ) ) // skip 'Exit' - isn't a theme
						rfeObjectData.selThemes.Add ( rfeObjectData.themes[ themeItemNum - 2 ] );
				}
			}
	
			break;
			
		default:
			break;
	}

	return 0;
}


int ResourceFileEditor::view ( Widget, XtPointer clientData )
{
	char textString [ EDITBUFSIZE ];
	int selectedItemNum;
	int count;

	selectedItemNum = menu ( "resourceFileEditor", "View data","Resource type",
								"Available,Selected,Exit,");
	count = 0;
								
	switch ( selectedItemNum )
	{
		case 1:
			{	// available items
				int availableItemNum;
				
				availableItemNum = menu ( "resourceFileEditor", "Available",
											"Resource","Applications,Themes,Aliases,Exit,");
				switch ( availableItemNum )
				{
					case 1:
						{	// show applications
							unsigned int length = 0;

							count = 0;
							textString [ 0 ] = 0;

							for ( count = 0; count < rfeObjectData.apps.Count(); count++ )
							{
								length += strlen ( rfeObjectData.apps[ count ].name );
								length += 1; // account for ','
							}
				
							length += 1;
	
							if ( length >= EDITBUFSIZE )
							{
								memError ( EDITBUFSIZE, "view/apps" );
								return 1;					
							}

							textString [ 0 ] = 0;
						
							for ( count = 0; count < rfeObjectData.apps.Count(); count++ )
							{
								strcat ( textString, rfeObjectData.apps[count].name ); 
								strcat ( textString, "\n" );
							}
							
							tablet ( "ResourceFileEditor", "available applications", textString );
						}
						
						break;
						
					case 2:
						{	//show themes
							unsigned int length = 0;

							count = 0;
							textString [ 0 ] = 0;

							for ( count = 0; count < rfeObjectData.themes.Count(); count++ )
							{
								length += strlen ( rfeObjectData.themes[ count ] );
								length += 1; // account for ','
							}
				
							length += 1;
	
							if ( length >= EDITBUFSIZE )
							{
								memError ( EDITBUFSIZE, "view/themes" );
								return 1;					
							}

							textString [ 0 ] = 0;
						
							for ( count = 0; count < rfeObjectData.themes.Count(); count++ )
							{
								strcat ( textString, rfeObjectData.themes[count] ); 
								strcat ( textString, "\n" );
							}
							
							tablet ( "ResourceFileEditor", "available themes", textString );
						}
						
						break;
						
					case 3:
						{	// show colour aliases					
							unsigned int length = 0;

							count = 0;
							textString [ 0 ] = 0;

							for ( count = 0; count < rfeObjectData.aliases.Count(); count++ )
							{
								length += strlen ( rfeObjectData.aliases[ count ] );
								length += 1; // account for ','
							}
				
							length += 1;
	
							if ( length >= EDITBUFSIZE )
							{
								memError ( EDITBUFSIZE, "view/aliases" );
								return 1;					
							}

							textString [ 0 ] = 0;
						
							for ( count = 0; count < rfeObjectData.aliases.Count(); count++ )
							{
								strcat ( textString, rfeObjectData.aliases[ count ] ); 
								strcat ( textString, "\n" );
							}
							
							tablet ( "ResourceFileEditor", "colour aliases", textString );
						}
					
					default:
						break;
				}														
			}
			
			break;

		case 2:
			{	// selected items
				int selectedItemNum;
				
				selectedItemNum = menu ( "resourceFileEditor", "Selected", "Resource",
											"Applications,Themes,Targets,Exit," );
				switch ( selectedItemNum )
				{
					case 1:
						{	// show selected applications
							unsigned int length = 0;

							count = 0;
							textString [ 0 ] = 0;

							for ( count = 0; count < rfeObjectData.selApps.Count(); count++ )
							{
								length += strlen ( rfeObjectData.selApps[ count ] );
								length += 1; // account for ','
							}
				
							length += 1;
	
							if ( length >= EDITBUFSIZE )
							{
								memError ( EDITBUFSIZE, "view/selected apps" );
								return 1;					
							}

							textString [ 0 ] = 0;
						
							for ( count = 0; count < rfeObjectData.selApps.Count(); count++ )
							{
								strcat ( textString, rfeObjectData.selApps[ count ] );
								strcat ( textString, "\n" );
							}
							
							tablet ( "ResourceFileEditor", "selected applications",textString );
						}
						
						break;
						
					case 2:
						{	// show selected themes
							unsigned int length = 0;

							count = 0;
							textString [ 0 ] = 0;

							for ( count = 0; count < rfeObjectData.selThemes.Count(); count++ )
							{
								length += strlen ( rfeObjectData.selThemes[ count ] );
								length += 1; // account for ','
							}
				
							length += 1;
	
							if ( length >= EDITBUFSIZE )
							{
								memError ( EDITBUFSIZE, "view/selected themes" );
								return 1;					
							}

							textString [ 0 ] = 0;

							for ( count = 0; count < rfeObjectData.selThemes.Count(); count++ )
							{
								strcat ( textString, rfeObjectData.selThemes[count] );
								strcat ( textString, "\n" );
							}
							
							tablet ( "ResourceFileEditor", "selected themes", textString );
						}
						
						break;
						
					case 3:
						{	// show selected targets

							unsigned int length = 0;

							count = 0;
							textString [ 0 ] = 0;

							for ( count = 0; count < rfeObjectData.opOn.Count(); count++ )
							{
								length += sizeof ( Ops );
								length += 22; // account for spaces, commas, number,...
							}
				
							length += 1;
	
							if ( length >= EDITBUFSIZE )
							{
								memError ( EDITBUFSIZE, "view/selected apps" );
								return 1;					
							}

							textString [ 0 ] = 0;
						
							for ( count = 0; count < rfeObjectData.opOn.Count(); count++ )
							{
								strcat ( textString, rfeObjectData.opOn[count].fileName );
								strcat ( textString, " " );
								strcat ( textString, rfeObjectData.opOn[count].resourceLabel );
								strcat ( textString, " " );
								strcat ( textString, rfeObjectData.opOn[count].returnedResource );
								strcat ( textString, ", " );									
								strcat ( textString, rfeObjectData.opOn[count].specifiedResource );
								strcat ( textString, "\n" );
							}
							
							tablet ( "ResourceFileEditor", "selected targets", textString );
						}
						
						break;
						
					default:
						break;
				}
			}
			
			break;
			
		default:
			break;
	}			

	return 0;
}


int ResourceFileEditor::exec ( Widget, XtPointer clientData )
{
	int count;
	int selectedItemNum;
	
	selectedItemNum = menu ( "resourceFileEditor", "Execute", "Action",
								"Delete,Edit,Replace,Undo,Exit," );

	switch ( selectedItemNum )
	{
		case 1:
			{	// delete - by number or from menu
				int delItemNum;
				
				delItemNum = menu ( "resourceFileEditor", "Delete", "Selected",
									"Application,Theme,Target,Undo,Exit," );
									
				switch ( delItemNum )
				{
					case 1:
						{	// delete a selected app
							int appItemNum;
							unsigned int length;
							int menuStart, menuEnd, numMenuItems;
				
							menuStart = 0;
							count = 0;

							if ( _maxAppMenuItems < rfeObjectData.selApps.Count() )
								numMenuItems = _maxAppMenuItems;
							else
								numMenuItems = rfeObjectData.selApps.Count();

							menuEnd = numMenuItems;				

							int keepLooping = 1;
			
							while ( keepLooping )
							{
								length = 0;
					
								for ( count = menuStart; count < numMenuItems; count++ )
								{
									length += strlen ( rfeObjectData.selApps[ count ] );
									length += 1; // account for ','
								}
				
								length += 15;	// and for 'All apps,' and 'Exit.\0'
				
								if ( length >= EDITBUFSIZE )
								{
									memError ( EDITBUFSIZE, "delete/apps" );
									return 1;					
								}
				
								currentEditLine [ 0 ] = 0;
					
								for ( count = menuStart; count < menuEnd; count++ )
								{			
									strcat ( currentEditLine, rfeObjectData.selApps[ count ] );
									strcat ( currentEditLine, "," );
								}

								if ( rfeObjectData.selApps.Count() > 1 )
									strcat ( currentEditLine, "All apps," );

								if ( menuEnd < rfeObjectData.selApps.Count() )
									strcat ( currentEditLine, "More," );
								else
									strcat ( currentEditLine, "Exit," );

								appItemNum = menu ( "resourceFileEditor", "App selection",
												"Applications", currentEditLine );

								currentEditLine [ 0 ] = 0;

								if ( appItemNum < ( numMenuItems + 2 ) )
								{
									appItemNum += menuStart;
									keepLooping = 0;
								}
								else
								{
									if ( ( appItemNum + menuStart ) == ( rfeObjectData.selApps.Count() + 2 ) )
									{
										keepLooping = 0;
										appItemNum = rfeObjectData.selApps.Count() + 2;
									}
				
									menuStart += numMenuItems;
									menuEnd += numMenuItems;
					
									if ( menuEnd > rfeObjectData.selApps.Count() )
									{
										menuStart = rfeObjectData.selApps.Count() - numMenuItems;
										menuEnd = rfeObjectData.selApps.Count();
									}
						
									if ( menuStart > ( rfeObjectData.selApps.Count() - numMenuItems ) )
									{
										menuStart = 0;
										menuEnd = numMenuItems;
									}
								}
							}	

							if ( rfeObjectData.selApps.Count() > 1 ) // more than just 'Exit' on menu
							{
								if ( ( appItemNum == ( rfeObjectData.selApps.Count() + 1 ) ) || ( appItemNum == ( numMenuItems + 1 ) ) )
									rfeObjectData.selApps.Release();
								else
									rfeObjectData.selApps.Remove ( rfeObjectData.selApps[ appItemNum - 1 ] );
							}
							else
								if ( appItemNum == 1 )
									rfeObjectData.selApps.Release();
						}
						
						break;

					case 2:
						{	// delete selected theme
							int themeItemNum;
							unsigned int length;
							int menuStart, menuEnd, numMenuItems;
				
							menuStart = 0;
							count = 0;

							if ( _maxThemeMenuItems < rfeObjectData.selThemes.Count() )
								numMenuItems = _maxThemeMenuItems;
							else
								numMenuItems = rfeObjectData.selThemes.Count();

							menuEnd = numMenuItems;				

							int keepLooping = 1;
			
							while ( keepLooping )
							{
								length = 0;
					
								for ( count = menuStart; count < numMenuItems; count++ )
								{
									length += strlen ( rfeObjectData.selThemes[ count ] );
									length += 1; // account for ','
								}
				
								length += 17;	// and for 'All themes,' and 'Exit.\0'
			
								if ( length >= EDITBUFSIZE )
								{
									memError ( EDITBUFSIZE, "delete/themes" );
									return 1;					
								}
				
								currentEditLine [ 0 ] = 0;
					
								for ( count = menuStart; count < menuEnd; count++ )
								{			
									strcat ( currentEditLine,rfeObjectData.selThemes[ count ] );
									strcat ( currentEditLine, "," );
								}

								if ( rfeObjectData.selThemes.Count() > 1 )
									strcat ( currentEditLine, "All themes," );

								if ( menuEnd < rfeObjectData.selThemes.Count() )
									strcat ( currentEditLine, "More," );
								else
									strcat ( currentEditLine, "Exit," );

								themeItemNum = menu ( "resourceFileEditor", "Select Target",
														"Themes", currentEditLine );

								currentEditLine [ 0 ] = 0;

								if ( themeItemNum < ( numMenuItems + 2 ) )
								{
									themeItemNum += menuStart;
									keepLooping = 0;
								}
								else
								{
									if ( ( themeItemNum + menuStart ) == ( rfeObjectData.selThemes.Count() + 2 ) )
									{
										keepLooping = 0;
										themeItemNum = rfeObjectData.selThemes.Count() + 2;
									}
				
									menuStart += numMenuItems;
									menuEnd += numMenuItems;
					
									if ( menuEnd > rfeObjectData.selThemes.Count() )
									{
										menuStart = rfeObjectData.selThemes.Count() - numMenuItems;
										menuEnd = rfeObjectData.selThemes.Count();
									}
						
									if ( menuStart > ( rfeObjectData.selThemes.Count() - numMenuItems ) )
									{
										menuStart = 0;
										menuEnd = numMenuItems;
									}
								}
							}	

							if ( rfeObjectData.selThemes.Count() > 1 ) // more than just 'Exit' on menu
							{
								if ( ( themeItemNum == ( rfeObjectData.selThemes.Count() + 1 ) ) || ( themeItemNum == ( numMenuItems + 1 ) ) )
									rfeObjectData.selThemes.Release();
								else
									rfeObjectData.selThemes.Remove ( rfeObjectData.selThemes[ themeItemNum - 1 ] );
							}
							else
								if ( themeItemNum == 1 )
									rfeObjectData.selThemes.Release();
						}
						
						break;
						
					case 3:
						{	// delete modification target
							int opItemNum;
							unsigned int length ;
							int menuStart, menuEnd, numMenuItems;
				
							menuStart = 0;
							count = 0;

							if ( myResources->maxMenuItems < rfeObjectData.opOn.Count() )
								numMenuItems = myResources->maxMenuItems;
							else
								numMenuItems = rfeObjectData.opOn.Count();

							menuEnd = numMenuItems;				

							int keepLooping = 1;
			
							while ( keepLooping )
							{
								length = 0;
					
								for ( count = menuStart; count < numMenuItems; count++ )
								{
									length += strlen ( rfeObjectData.opOn[ count ].fileName );
									length += strlen ( rfeObjectData.opOn[ count ].resourceLabel );
									length += strlen ( rfeObjectData.opOn[ count ].returnedResource );
									length += 3; // account for 2 spaces and ','
								}
				
								length += 18;	// and for 'All targets,' and 'Exit,\0'
			
								if ( length >= EDITBUFSIZE )
								{
									memError ( EDITBUFSIZE, "delete/target" );
									return 1;					
								}
				
								currentEditLine [ 0 ] = 0;
					
								for ( count = menuStart; count < menuEnd; count++ )
								{		
									strcat ( currentEditLine, rfeObjectData.opOn[ count ].fileName );
									strcat ( currentEditLine, " " );
									strcat ( currentEditLine, rfeObjectData.opOn[ count ].resourceLabel );
									strcat ( currentEditLine, " " );
									strcat ( currentEditLine, rfeObjectData.opOn[ count ].returnedResource );
									strcat ( currentEditLine, "," );
								}

								if ( rfeObjectData.opOn.Count() > 1 )
									strcat ( currentEditLine, "All targets," );

								if ( menuEnd < rfeObjectData.opOn.Count() )
									strcat ( currentEditLine, "More," );
								else
									strcat ( currentEditLine, "Exit," );

								opItemNum = menu ( "resourceFileEditor", "Select Target",
													"Resources", currentEditLine );

								currentEditLine [ 0 ] = 0;

								if ( opItemNum < ( numMenuItems + 2 ) )
								{
									opItemNum += menuStart;
									keepLooping = 0;
								}
								else
								{
									if ( ( opItemNum + menuStart ) == ( rfeObjectData.opOn.Count() + 2 ) )
									{
										keepLooping = 0;
										opItemNum = rfeObjectData.opOn.Count() + 2;
									}
				
									menuStart += numMenuItems;
									menuEnd += numMenuItems;
					
									if ( menuEnd > rfeObjectData.opOn.Count() )
									{
										menuStart = rfeObjectData.opOn.Count() - numMenuItems;
										menuEnd = rfeObjectData.opOn.Count();
									}
						
									if ( menuStart > ( rfeObjectData.opOn.Count() - numMenuItems ) )
									{
										menuStart = 0;
										menuEnd = numMenuItems;
									}
								}
							}	
							
							if ( rfeObjectData.opOn.Count() > 1 ) // more than just 'Exit' on menu
							{
								if ( ( opItemNum == ( rfeObjectData.opOn.Count() + 1 ) ) || ( opItemNum == ( numMenuItems + 1 ) ) )
									rfeObjectData.opOn.Release();
								else
									rfeObjectData.opOn.RemoveItemAt ( opItemNum - 1 );
							}
							else
								if ( opItemNum == 1 )
									rfeObjectData.opOn.Release();

						}	// end case 3: delete located resource
						
						break;
						
					case 4:
						{	// delete undo files
							char textString [ MAXCHARS ];
							
							if ( _undoFileNum == 0 )
							{
								strcpy ( textString, "No undo files have been saved in this session.\n" );
								strcat ( textString, "Delete files from any previous sessions?" );
								
								tablet ( "ResourceFileEditor", "Delete undo files", textString );
								
								if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
								{
									currentEditLine [ 0 ] = 0;
									return 1;
								}
								
								currentEditLine [ 0 ] = 0;
							}	
							
							strcpy ( textString, "rm -f " );
							strcat ( textString, _homeDir );
							strcat ( textString, "/" );
							strcat ( textString, myResources->resourceDir );
							strcat ( textString, "/" );
							strcat ( textString, "*undo" );
							sysCall ( textString );
							textString [ 0 ] = 0;
						}
						
						break;
							
					default:
						break;
				}	// end switch delItemNum
			}
				
			break;	// end case 1: delete
			
		case 2:	// edit functions
			{	// edit - choose file - selected files by menu button
				int editItemNum;
				
				editItemNum = menu ( "resourceFileEditor", "Edit files", currentThemeName,
									"Applications,Resources,Configuration,Exit," );
										
				switch ( editItemNum )
				{
					case 1:
						{	// edit applications
							int appItemNum;
							unsigned int length;
							int menuStart, menuEnd, numMenuItems;
	
							menuStart = 0;
							count = 0;
							currentEditLine [ 0 ] = 0;	// something hasn't cleared this after use...

							if ( _maxAppMenuItems < rfeObjectData.apps.Count() )
								numMenuItems = _maxAppMenuItems;
							else
								numMenuItems = rfeObjectData.apps.Count();
										   
							menuEnd = numMenuItems;

							int keepLooping = 1;
									   
							while ( keepLooping )
							{			
								length = 0;

								for ( count = menuStart; count < numMenuItems; count++ )
								{
									length += strlen ( rfeObjectData.apps[ count ].name );
									length += 1; // account for ','
								}

								length += 6;	// add for 'more,' or 'Exit.\0'
										   
								if ( length >= EDITBUFSIZE )
								{
									memError ( EDITBUFSIZE, "edit/apps" );
									return 1;					
								}

								for ( count = menuStart; count < menuEnd; count++ )
								{			
									strcat ( currentEditLine, rfeObjectData.apps[ count ].name );
									strcat ( currentEditLine, "," );
								}

								if ( menuEnd < rfeObjectData.apps.Count() )
									strcat ( currentEditLine, "More," );
								else
									strcat ( currentEditLine, "Exit," );

								appItemNum = menu ( "resourceFileEditor", "Edit resource",
													"Applications", currentEditLine );

								currentEditLine [ 0 ] = 0;

								if ( appItemNum < ( numMenuItems + 1 ) )
								{
									appItemNum += menuStart;
									keepLooping = 0;
								}
								else
								{
									if ( ( appItemNum + menuStart ) == ( rfeObjectData.apps.Count() + 1 ) )
									{
										keepLooping = 0;
										appItemNum = rfeObjectData.apps.Count() + 2;
									}

									menuStart += numMenuItems;
									menuEnd += numMenuItems;
							
									if ( menuEnd > rfeObjectData.apps.Count() )
									{
										menuStart = rfeObjectData.apps.Count() - numMenuItems;
										menuEnd = rfeObjectData.apps.Count();
									}

									if ( menuStart > ( rfeObjectData.apps.Count() - numMenuItems ) )
									{
										menuStart = 0;
										menuEnd = numMenuItems;
									}
								}

								currentEditLine [ 0 ] = 0;

							}	// end while keepLooping

							// edit selected file
							if ( appItemNum < ( rfeObjectData.apps.Count() + 2 ) )
							{	// read file, edit, write file
								FILE *fileHandle;
								char workingFile [ MAXCHARS ];

								strcpy ( workingFile, _homeDir );
								strcat ( workingFile, "/" );
								strcat ( workingFile, myResources->appResDir );
								strcat ( workingFile, "/" );
								strcat ( workingFile, rfeObjectData.apps[ appItemNum - 1].resourceName );
								strcat ( workingFile, "-" );
								strcat ( workingFile, currentThemeName );
								
								char * textString;

								if ( ( fileHandle = fopen ( workingFile, "r" ) ) != NULL )
								{
									long length;
									fseek ( fileHandle, 0L, SEEK_END );
									length = ftell ( fileHandle );

									textString = new char [ length + 1 ];
									fseek ( fileHandle, 0L, SEEK_SET );
									fread ( textString, length, 1, fileHandle );
									fclose ( fileHandle );
									tablet ( "ResourceFileEditor", workingFile, textString );
		
									if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
													currentEditLine [ 0 ] = 0;
									else
									{
										length = strlen ( currentEditLine );
			
										if ( ( fileHandle = fopen ( workingFile, "w" ) ) != NULL )
											fwrite ( currentEditLine, 1, length, fileHandle );
										else
											tablet ( "ResourceFileEditor_error", "edit",
													 "edit/avaliable/apps: file write failed\n" );
				
										fclose ( fileHandle );
										currentEditLine [ 0 ] = 0;
									}

									delete[] textString;
								}
								else
								{
									strcpy ( currentEditLine, workingFile );
									strcat ( currentEditLine, "\nCan't read file.\n" );
									tablet ( "ResourceFileEditor_error", "file error", currentEditLine );
									currentEditLine [ 0 ] = 0;
									return 1;
								}
							}	// end if appItemNum
						} // end case 1: edit apps

						break;

					case 2:
						{	// edit resources
							int resItemNum;
							FILE *fileHandle;
							char workingFile [ MAXCHARS ];
						
							strcpy ( workingFile, _homeDir );
							strcat ( workingFile, "/" );

							char * textString;
			
							resItemNum = menu ( "resourceFileEditor", "Edit resource",
												"Applications", "Xdefaults,Xresources,Gtkrc,Menu,Style,Exit," );

							switch ( resItemNum )
							{
								case 1:
									{	// .Xdefaults, .Xresources
										strcat ( workingFile, myResources->XdefaultsDir );
										strcat ( workingFile, "/.Xdefaults-" );
										strcat ( workingFile, currentThemeName );
									}
								
									break;
							
								case 2:
									{
										strcat ( workingFile, myResources->XdefaultsDir );
										strcat ( workingFile, "/.Xresources-" );
										strcat ( workingFile, currentThemeName );
										
									}							
									
									break;

								case 3:
									{
										strcat ( workingFile, myResources->XdefaultsDir );
										strcat ( workingFile, "/gtk/gtkrc-" );
										strcat ( workingFile, currentThemeName );
									}							
									
									break;

								case 4:
									{	// Menu
										strcat ( workingFile, myResources->menusDir );
										strcat ( workingFile, "/" );
										strcat ( workingFile, currentThemeName );
										strcat ( workingFile, ".menu" );
									}
								
									break;

								case 5:
									{	// Style
										strcat ( workingFile, myResources->stylesDir );
										strcat ( workingFile, "/" );
										strcat ( workingFile, currentThemeName );
									}
								
									break;
								
								default:
									break;
									
							}	// end switch resItemNum

							if ( resItemNum < 6 )
							{	// read file, edit, write file
								if ( ( fileHandle = fopen ( workingFile, "r" ) ) != NULL )
								{
									long length;
									fseek ( fileHandle, 0L, SEEK_END );
									length = ftell ( fileHandle );

									textString = new char [ length + 1 ];
									fseek ( fileHandle, 0L, SEEK_SET );
									fread ( textString, length, 1, fileHandle );
									fclose ( fileHandle );
									tablet ( "ResourceFileEditor", workingFile, textString );
		
									if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
										currentEditLine [ 0 ] = 0;
									else
									{
										length = strlen ( currentEditLine );
			
										if ( ( fileHandle = fopen ( workingFile, "w" ) ) != NULL )
											fwrite ( currentEditLine, 1, length, fileHandle );
										else
											tablet ( "ResourceFileEditor_error", "file error",
														"edit/avaliable/apps: file write failed\n" );
				
										fclose ( fileHandle );
										currentEditLine [ 0 ] = 0;
									}

									delete[] textString;
								}
								else
								{
									strcpy ( currentEditLine, workingFile );
									strcat ( currentEditLine, "\nCan't read file.\n" );
									tablet ( "ResourceFileEditor_error", "file error", currentEditLine );
									currentEditLine [ 0 ] = 0;
									return 1;
								}
							}	// end if not resItemNum = 'Exit'
						}	// end edit resources
					
						break;

					case 3:
						{	// edit blackhome configuration files
							int configItemNum;
							char menuString [ MAXCHARS ];
							char workingFile [ MAXCHARS ];
							char * textString;
							FILE * fileHandle;
														
							strcpy ( menuString, "AppDefsList,LinkDefsList,User Spec," );
							strcat ( menuString, "Colour alias,App alias,RFE rc,Exit," );

							configItemNum = menu ( "resourceFileEditor_select",
													"Edit resource", "Configuration", menuString );

							strcpy ( workingFile, _homeDir );
							strcat ( workingFile, "/" );
							strcat ( workingFile, myResources->resourceDir );
							strcat ( workingFile, "/" );
						
							switch ( configItemNum )
							{
								case 1:	strcat ( workingFile, "appDefsList" ); break;
								case 2: strcat ( workingFile, "linkDefsList" ); break;
								case 3: strcat ( workingFile, "userSpec" ); break;
								case 4: strcat ( workingFile, ".colourAliasList" ); break;
								case 5: strcat ( workingFile, ".appNameAlias" ); break;
								case 6: strcat ( workingFile, ".resourceFileEditorrc" ); break;
								default: break;
							}

							if ( configItemNum < 7 )
							{	// read file, edit, write file
								if ( ( fileHandle = fopen ( workingFile, "r" ) ) != NULL )
								{
									long length;
									fseek ( fileHandle, 0L, SEEK_END );
									length = ftell ( fileHandle );

									textString = new char [ length + 1 ];
									fseek ( fileHandle, 0L, SEEK_SET );
									fread ( textString, length, 1, fileHandle );
									fclose ( fileHandle );
									tablet ( "ResourceFileEditor", workingFile, textString );
		
									if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
										currentEditLine [ 0 ] = 0;
									else
									{
										length = strlen ( currentEditLine );
			
										if ( ( fileHandle = fopen ( workingFile, "w" ) ) != NULL )
											fwrite ( currentEditLine, 1, length, fileHandle );
										else
											tablet ( "ResourceFileEditor_error", "file error",
														"edit/avaliable/apps: file write failed\n" );
				
										fclose ( fileHandle );
										currentEditLine [ 0 ] = 0;
									}

									delete[] textString;
								}
								else
								{
									strcpy ( currentEditLine, workingFile );
									strcat ( currentEditLine, "\nCan't read file.\n" );
									tablet ( "ResourceFileEditor_error", "file error", currentEditLine );
									currentEditLine [ 0 ] = 0;
									return 1;
								}
							}	// end if not configItemNum = 'Exit'
						}
						
						break;

					default:
						break;

				}	// end switch editItemNum
			}	// end edit functions
			
			break;

		case 3:
			{	// replace targets with selections ( any of colour, widget, ... )
				char textString [ MAXCHARS ];

				if ( rfeObjectData.opOn.Count() == 0 )
				{
					snprintf ( textString, sizeof ( textString ), "No targets - use 'Find' to specify a search item." );
					tablet ( "ResourceFileEditor_error", "Error", textString );
					return 1;
				}

				char * newResource;
				char specifiedResource [ MAXCHARS ];
				char newColour [ MAXCHARS ];
				char substituteColour [ MAXCHARS ];
				int error;							
				int keepLooping;
				int prompt;

				cleanReplace();
				ColourSpecConvert *myConvert = new ColourSpecConvert;
				error = myConvert->setRgbTxtPath ( myResources->rgbTxtDir );

				if ( error )
					reportColourError ( myResources->rgbTxtDir, error );

				myConvert->setfReturnSize ( _fReturnSize );
				myConvert->setReturnSize ( _returnSize );
		
				strcpy ( textString, _homeDir );
				strcat ( textString, "/" );
				strcat ( textString, myResources->rgbTxtDir );

				error = myConvert->setResourceFilePath ( textString );

				if ( error )
				{
					strcat ( textString, "Couldn't access file .colourAliasList.\n" );
					tablet ( "ResourceFileEditor_error", "file error", textString );
					delete myConvert;
					return 1;
				}

				specifiedResource [ 0 ] = 0;
				prompt = 1;	// cleared after 1st input if _useFirstResource

				// get replacement resources
				for ( count = 0; count < rfeObjectData.opOn.Count(); count++ )
				{
					error = 0;

					if ( strcmp ( rfeObjectData.opOn[ count ].returnedResource, specifiedResource ) != 0 )
					{	// prompt for new one
						strcpy ( specifiedResource, rfeObjectData.opOn[ count ].returnedResource );
						keepLooping = 1;

						while ( keepLooping && prompt )
						{	// loop until replacement resource spec verifies ok
							strcpy ( textString, "replace " );
							strcat ( textString, rfeObjectData.opOn[ count ].returnedResource );
							strcat ( textString, " with..." );
							tablet ( "Find_resource", textString, "" );
							textString [ 0 ] = 0;

							if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
							{
								keepLooping = 0;
								error = 1;							
								count = rfeObjectData.opOn.Count();
							}

							if ( strlen ( currentEditLine ) >= MAXCHARS )
							{
								currentEditLine [ 0 ] = 0;
								memError ( MAXCHARS, "exec/replace" );
								keepLooping = 0;
								error = 1;							
								count = rfeObjectData.opOn.Count();
							}

							strcpy ( textString, currentEditLine );
							currentEditLine [ 0 ] = 0;
							
							// check resource type
							if ( ( ( int ) rfeObjectData.opOn[ count ].resourceType == 2 ) && ( !error ) ) // was an X colour specified?
							{	// check colour resource
								error = myConvert->setColour ( textString );

								if ( error )
								{
									reportColourError ( textString, error );
									currentEditLine [ 0 ] = 0;
								}
								else
									reportColour ( myConvert);

								if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
								{
									keepLooping = 0;
									error = 1;							
									count = rfeObjectData.opOn.Count();
								}
								
								currentEditLine [ 0 ] = 0;
								strcpy ( newColour, textString);																
							}	// end check for colour resource

							// check here for other type, app, widget etc							

							if ( !error )	// specified resource valid
								keepLooping = 0;
								
							if ( _useFirstResource )
								prompt = 0;
						}	// end while keepLooping

						if ( error )
						{
							delete myConvert;
							return ( error );
						}

						// found valid resource - check for resourceType
						if ( ( int ) rfeObjectData.opOn[ count ].resourceType == 2 )
						{	// was colour resource?
							strcpy ( substituteColour, newColour ); 	// use if resource file colour is same as previously

							// return shade in currentEditLine
							if ( _shadeReplace )
							{
								shadeColour ( rfeObjectData.opOn[ count ].returnedResource, newColour );
								strcpy ( substituteColour, currentEditLine );
								currentEditLine [ 0 ] = 0;
							}

							strcpy ( textString, formatColour ( substituteColour ) );

							if ( strlen ( textString ) == 0 )
							{
								tablet ( "ResourceFileEditor_error", "format colour failed", substituteColour );
								delete myConvert;
								return 1;	// didn't get a formatted colour back
							}

							// force rgb spec on blackbox menus
							if ( strstr ( rfeObjectData.opOn[ count ].fileName, ".menu" ) != NULL )
								strcpy ( textString, myConvert->rgbColour() );

							// force # spec on Eterm app-defaults files - takes ints as well, don't know mag limits
							if ( strstr ( rfeObjectData.opOn[ count ].fileName, "Eterm-color" ) != NULL )
								strcpy ( textString, myConvert->hashColour() );
	
							newResource = new char [ strlen ( textString ) + 1 ];
							strcpy ( newResource, textString );
							rfeObjectData.replace.Add ( newResource );
						}	// end check colour resource
					}
					else
					{	// found a similar instance of specified resource, don't prompt for input
						if ( ( int ) rfeObjectData.opOn[ count ].resourceType == 2 )
						{	// was colour resource?
							if ( _shadeReplace )
							{
								shadeColour ( rfeObjectData.opOn[ count ].returnedResource, newColour );
								strcpy ( substituteColour, currentEditLine );
								currentEditLine [ 0 ] = 0;
							}
							
							strcpy ( textString, formatColour ( substituteColour ) );

							if ( strlen ( textString ) == 0 )
							{
								tablet ( "ResourceFileEditor_error", "format colour failed", substituteColour );
								delete myConvert;
								return 1;	// didn't get a formatted colour back
							}

							// force rgb spec on blackbox menus
							if ( strstr ( rfeObjectData.opOn[ count ].fileName, ".menu" ) != NULL )
								strcpy ( textString, myConvert->rgbColour() );

							// force # spec on Eterm app-defaults files
							if ( strstr ( rfeObjectData.opOn[ count ].fileName, "Eterm-color" ) != NULL )
								strcpy ( textString, myConvert->hashColour() );

							newResource = new char [ strlen ( textString ) + 1 ];
							strcpy ( newResource, textString );
							rfeObjectData.replace.Add ( newResource );
						}	// end check colour resource
					}
				}	// end get new resources to replace old ones

				error = substituteResource();	// do new resources for old

				if ( error )
				{
					sprintf ( textString, "%d\n", error );					
					strcat ( textString, "error: Resource replacement failed\n" );
					tablet ( "ResourceFileEditor_error", "Exec/Replace", textString );
					currentEditLine [ 0 ] = 0;
					delete myConvert;
					return error;
				}
				
				FILE *fileHandle;
				char workingFile [ MAXCHARS ];			

				if ( _saveResourceMods )
				{
					char numString [ MAXCHARS ];

					_undoFileNum += 1;	// append to undo file name for extra ID
					strcpy ( workingFile, _homeDir );
					strcat ( workingFile, "/" );
					strcat ( workingFile, myResources->resourceDir );
					strcat ( workingFile, "/" );
					strcat ( workingFile, currentThemeName );
					strcat ( workingFile, "-" );
					sprintf ( numString, "%u", _undoFileNum );
					strcat  ( numString, "\0" );
					strcat ( workingFile, numString );
					strcat ( workingFile, ".undo" );

					if ( ( fileHandle = fopen ( workingFile, "r" ) ) != NULL )
					{
						tablet ( "ResourceFileEditor", "Undo file exists", workingFile );

						if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
						{
							workingFile [ 0 ] = 0;
							error = 1;
						}
						else
						{
							strcpy ( workingFile, currentEditLine );
							error = 0;
						}

						fclose ( fileHandle );
					}						

					currentEditLine [ 0 ] = 0;
					char title [ strlen ( specifiedResource ) + 9 ];
					strcpy ( title, "Replace " );
					strcat ( title, specifiedResource );
					
					tablet ( "ResourceFileEditor", title, "Press Ok and wait...\nor Cancel to quit replace" );
	
					if ( strcmp (currentEditLine, "tabletCancel" ) == 0 )
					{
						currentEditLine[0] = 0;
						delete myConvert;
						return 1;
					}

					currentEditLine [ 0 ] = 0;

					if ( ( fileHandle = fopen ( workingFile, "w" ) ) != NULL )
					{					
						textString [ 0 ] = 0;
						numString [ 0 ] = 0;
						
						if ( !error )
						{
							for ( count = 0; count < rfeObjectData.opOn.Count(); count++ )
							{
								sprintf ( numString, "%u", ( unsigned int ) rfeObjectData.opOn[ count ].resourceType );
								strcpy ( textString, numString );
								strcat ( textString, " " );
								strcat ( textString, rfeObjectData.opOn[ count ].fileName );
								strcat ( textString, " " );
								strcat ( textString, rfeObjectData.opOn[ count ].resourceLabel );
								strcat ( textString, " " );
								strcat ( textString, rfeObjectData.opOn[ count ].returnedResource );
								strcat ( textString, " " );
								strcat ( textString, rfeObjectData.replace[ count ] );
								strcat ( textString, "\n" );
								fwrite ( textString, 1, strlen ( textString ), fileHandle );
							}

							fflush ( fileHandle );
							fclose ( fileHandle );
						}
						else
						{
							strcpy ( textString, "Cannot write undo file\n" );
							strcat ( textString, workingFile );
							tablet ( "ResourceFileEditor_error", "resource undo write error", textString );
							currentEditLine [ 0 ] = 0;					
						}
					}
				}

				textString [ 0 ] = 0;
				workingFile [ 0 ] = 0;
				delete myConvert;

				// finally, if the 'execute app' switch is on AND the app is on the selected list
				if ( _execOnModify )
				{
					error = execOnModify();

					if ( error )
					{
						tablet ( "ResourceFileEditor_error", "Exec/Replace", "Execute on modify failed." );
						currentEditLine [ 0 ] = 0;
						return 1;
					}
				}
			}	// end case 3: replace

			break;
			
		case 4:
			{	// undo, theoretically.
				char textString [ MAXCHARS ];
				char workingFile[ MAXCHARS ];
				unsigned int resourceType;
				char * newResource;
				char fileName [ MAXCHARS ];
				char resourceLabel [ MAXCHARS ];
				char returnedResource [ MAXCHARS ];
				char replacementResource [ MAXCHARS ];
				char numString [ MAXCHARS ];
				int error, undoItemNum;
				bool isGnomeColour;

				FILE * stream;
				error = 0;

				undoItemNum = menu ( "resourceFileEditor_select", "Undo edit",
										"Undo type", "Last,Previous,Exit," );
				switch ( undoItemNum )
				{
					case 1:
						{	// undo last
							if ( _undoFileNum == 0 )
							{
								snprintf ( textString, sizeof ( textString ), "Nothing in \
current session to undo" );
								tablet ( "ResourceFileEditor_error", "Error", textString );
								return 1;
							}

							strcpy ( workingFile, _homeDir );
							strcat ( workingFile, "/" );
							strcat ( workingFile, myResources->resourceDir );
							strcat ( workingFile, "/" );
							strcat ( workingFile, currentThemeName );
							strcat ( workingFile, "-" );
							sprintf ( numString, "%u", _undoFileNum );
							strcat ( workingFile, numString );
							strcat ( workingFile, ".undo" );
						}
						
						break;
						
					case 2:
						{	// undo previous
							strcpy ( textString, "Enter undo filename" );
							tablet ( "ResourceFileEditor", textString, "" );

							if ( strcmp ( currentEditLine, "tabletCancel" ) == 0 )
							{
								currentEditLine [ 0 ] = 0;
								textString [ 0 ] = 0;
								workingFile [ 0 ] = 0;
								return 1;
							}

							strcpy ( workingFile, _homeDir );
							strcat ( workingFile, "/" );
							strcat ( workingFile, myResources->resourceDir );
							strcat ( workingFile, "/" );
							strcat ( workingFile, currentEditLine );
							currentEditLine [ 0 ] = 0;
							textString [ 0 ] = 0;
						}
						
						break;
						
					default:
						return 0;
				}	// end switch undoItemNum

				stream = fopen ( workingFile, "r" );

				if ( !stream )
				{
					strcpy ( textString, "Exec/undo: cannot read\n");
					strcat ( textString, workingFile );
					tablet ( "ResourceFileEditor_error", "file error", textString );
					currentEditLine [ 0 ] = 0;
					textString [ 0 ] = 0;
					return 1;
				}

				Ops tempOps;
				cleanOperate();
				cleanReplace();

				while ( !feof ( stream ) )
				{
					resourceType = 0;
					fileName[0] = 0;

					fscanf ( stream, "%u %s", &resourceType, fileName );
					isGnomeColour = FALSE;	// set for every new resource exchange in undo file

					if ( strstr ( fileName, ".menu" ) == NULL )
					{
						if ( strstr ( fileName, "gtkrc-" ) != NULL )
						{
							// if there's an rgbi colour it's been substituted by RFE
							// label starts at end of resourceType: where's this?
							int ptr = ftell ( stream );	// start of file resourceLabel
							char c = 0;

							count  = 0;
							textString [ 0 ] = 0;

							while ( c != 0x0a )
							{
								fread ( &c, 1, 1, stream );
								textString [ count ] = c;
								count++;
							}

							textString [ count ] = 0;

							if ( strstr ( textString, "rgbi:" ) != NULL )
							{
								isGnomeColour = TRUE;
								count = strstr ( textString, "rgbi:" ) - textString;
								ptr = 0;

								while ( ptr < count )
								{
									resourceLabel [ ptr ] = textString [ ptr ];
									ptr++;
								}

								resourceLabel [ count ] = 0;
//if ( resourceType == 2	)	// colour resource
//{
								ptr = 0;

								while ( textString [ count ] != ' ' )	// wsp after rgbi spec
								{
									returnedResource [ ptr  ] = textString [ count ];
									count++;
									ptr++;
								}

								count++;	// advance past wsp to replacement
								returnedResource [ ptr ] = 0;
								ptr = 0;

								while ( textString [ count ] != 0x0a )
								{
									replacementResource [ ptr ] = textString [ count ];
									count++;
									ptr++;
								}

								replacementResource [ ptr ] = 0;
//}
							}
						}


						if ( isGnomeColour == FALSE ) // didn't get an rgbi spec for gnome colour
							error = fscanf ( stream, "%s %s %s\n", resourceLabel, returnedResource,
											replacementResource );
					}
					else	// blackhome menu file
					{	// another exception for multiple resource entries per line
						char c = 0;
						unsigned int count, end;

						count = 0;

						while ( c != 0x0a )
						{
							fread ( &c, 1, 1, stream );
							textString [ count ] = c;
							count++;
						}

						textString [ count ] = 0;
						end = strlen ( textString) - 1;
						count = 0;
						// first strip trailing wsp
						while ( ( textString [ end ] == ' ' ) || ( textString [ end ] == '\t' ) )
							end--;

						int ignore = 0;
  
						while ( !ignore )
						{
							c = textString [ end ];

							if ( c == ' ' )
								ignore = 1;

							if ( c == '\t' )
								ignore = 1;

							end--;
						}	
						// found end word ( replacementResource )
						end--;

						while ( ignore )
						{
							c = textString [ end ];

							if ( c != ' ' )
								ignore = 0;

							if ( c != '\t' )
								ignore = 0;

							end--;
						}
						// skip intervening wspace	
						end--;

						while ( !ignore )
						{
							c = textString [ end ];

							if ( c == ' ' )
								ignore = 1;

							if ( c == '\t' )
								ignore = 1;

							end--;
						}
						// found penultimate word ( returnedResource )
						end--;

						while ( ignore )
						{
							c = textString [ end ];
							if ( c != ' ' )

								ignore = 0;

							if ( c != '\t' )
								ignore = 0;

							end--;
						}
						// now at end of 'resourceLabel'

						end++;	// back up 1 to keep last !wsp char
						count = 0;

						while ( count <= end )
						{
							resourceLabel [ count ] = textString [ count ];
							count++;
						}

						resourceLabel [ count ] = 0;
						count ++;	// skip past last 'label' char to next wsp

						int tPtr = 0;

						while ( count < strlen ( textString ) )
						{
							numString [ tPtr ] = textString [ count ];
							count++;
							tPtr++;
						}

						numString [ tPtr ] = 0;
						sscanf ( numString, "%s %s", returnedResource, replacementResource );
						numString [ 0 ] = 0;
						textString [ 0 ] = 0;
					}

					if ( strlen ( fileName ) != 0 )
					{
						tempOps.fileName = new char [ strlen ( fileName ) + 1 ];
						tempOps.resourceLabel = new char [ strlen ( resourceLabel ) + 1 ];
						tempOps.returnedResource = new char [ strlen ( replacementResource ) + 1 ];
						tempOps.specifiedResource = new char [ strlen ( returnedResource ) + 1 ];
						tempOps.resourceType = new unsigned int;

						strcpy ( tempOps.fileName, fileName );
						strcpy ( tempOps.resourceLabel, resourceLabel );
						strcpy ( tempOps.returnedResource, replacementResource );
						strcpy ( tempOps.specifiedResource, returnedResource ); // same since this is a replay...
						tempOps.resourceType = (unsigned int * ) resourceType; // signifies 'colour'
						rfeObjectData.opOn.Add ( tempOps );
						newResource = new char [ strlen ( returnedResource ) + 1 ];
						strcpy ( newResource, returnedResource );
						rfeObjectData.replace.Add ( newResource );
					}
				}

				fclose ( stream );
				error = substituteResource();	// do old resources for new

				if ( error )
				{
					sprintf(numString, "%d", error );				
					strcpy ( textString, "Undo error: " );
					strcat ( textString, numString );
					strcat ( textString, "\nOperation aborted - check undo file:\n");
					strcat ( textString, workingFile );
					tablet ( "ResourceFileEditor_error", "substitute resource failed", textString );
					currentEditLine [ 0 ] = 0;
					textString [ 0 ] = 0;
					return error;
				}

				// remove undo file
				char cmd [ strlen ( workingFile ) + 4 ];
				strcpy ( cmd, "rm " );
				strcat ( cmd, workingFile );
				error = sysCall ( cmd );

				if ( error == 0 )
					_undoFileNum--;

				if ( _execOnModify )
				{
					error = execOnModify();

					if ( error )
					{
						tablet ( "ResourceFileEditor_error", "Exec/Undo", "Execute on modify failed." );
						currentEditLine [ 0 ] = 0;
						return 1;
					}
				}

				// last replacement list colours no longer exist after undo so
				// a subsequent 'Exec/replace' will fail. Delete list.
				cleanOperate();
			}	// end case 4: exec/undo
			
			break;

		default:
			break;
	}	// end switch selectedItemNum

	return 0;
}


// utility functions

char * ResourceFileEditor::currentTheme ( char * _homeDir, void * resources )
{	// search for unique string currentThemeSpecifier identifying line in
	// default.menu containing name of current theme

	rfeResources * myResources = ( rfeResources * ) resources;
	FILE * fileHandle;
	unsigned int count, themeCount;
	int lineCount;
	char c;
	char workingFile[ MAXCHARS ];
	char textString [ MAXCHARS ];

	if ( strlen ( myResources->currentThemeSpecifier ) == 0 )
		return "blackhome";

	strcpy ( workingFile, _homeDir );
	strcat ( workingFile, "/" );
	strcat ( workingFile, myResources->menusDir );
	strcat ( workingFile, "/default.menu" );

	if ( ( fileHandle = fopen ( workingFile, "r" ) ) != NULL )
	{ 
		c = 0;
		lineCount = 0;

		do
		{
			fread ( &c, 1, 1, fileHandle );    
			textString [ lineCount ] = c;

			if ( c == 0x0a ) // got current line
			{
				if ( strstr ( textString, myResources->currentThemeSpecifier ) != NULL )
				{
					count = 0;
				
					while (textString[count] != '{')
						count++; // found start of theme name entry

					count++;
					themeCount = 0;

					while (textString[count] != '}' )
					{
						currentThemeName[themeCount] = textString[count];
						themeCount++;
						count++;
					
						if (count >= MAXCHARS)
						{
							printf("theme name too long, rejecting\n");
							return "";
						}
					}

					currentThemeName[themeCount] = 0x00;
					return ( char * ) currentThemeName;
				}	
				else
				{
					lineCount = -1;
					textString [ 0 ] = 0;
				}
			}  // end got current line

			lineCount++;

		}  while ( !feof ( fileHandle ) );

		fclose ( fileHandle );
	}
	else
	{
		strcpy ( textString, "ResourceFileEditor::currentTheme\n" );
		strcat ( textString, "Cannot read " );
		strcat ( textString, workingFile );
		strcat ( textString, "\nUnable to identify current theme name\n" );
		tablet ( "ResourceFileEditor_error", "file error", textString );
	}

	return "No theme";
}


int ResourceFileEditor::listThemes ( char * _homeDir, void * resources )
{	
	FILE *fileHandle;
	char workingFile [ MAXCHARS ];
	char textString [ MAXCHARS ];
	char themeName [ MAXCHARS ];
	int count, themeCount;
	char c;
	char * temp;
	rfeResources * myResources = ( rfeResources * ) resources;

	strcpy ( workingFile, _homeDir );
	strcat ( workingFile, "/" );
	strcat ( workingFile, myResources->stylesDir );
	strcat ( workingFile, "/stylesmenu" );

	if ( ( fileHandle = fopen ( workingFile, "r" ) ) != NULL )
	{
		c = 0;
		count = 0;
		themeCount = 0;
		
		do
		{
			fread ( &c, 1, 1, fileHandle );
			textString [ count ] = c;

			if ( c == 0x0a ) // got current line
			{			
				textString [ count ] = 0;

				if ( strlen ( textString ) != 0 )
				{
					if ( textString [ 0 ] != '#' )
					{	// reject comments
						themeCount = 0;

						while ( textString [ themeCount ] != '(' )
							themeCount++;

						themeCount++;
						int ptr = 0;
						
						while ( textString [ themeCount ] != ')' )
						{
							themeName [ ptr ] = textString [ themeCount ];
							themeCount++;
							ptr++;
						}
					
						themeName [ ptr ] = 0;
						temp = new char [ strlen ( themeName ) + 1 ];
						strcpy ( temp, themeName );
						rfeObjectData.themes.Add ( temp );
					}
				}
				
				count = -1;
			}
			
			count++;
		
		}  while ( !feof ( fileHandle ) );

		fclose ( fileHandle );
	}
	else
	{
		strcpy ( textString, "ResourceFileEditor::listThemes\n" );
		strcat ( textString, "Cannot read " );
		strcat ( textString, workingFile );
		strcat ( textString, "\nUnable to construct theme list\n" );
		tablet ( "ResourceFileEditor_error", "file error", textString );
		return 1;
	}

	return 0;
}


int ResourceFileEditor::listApps ( char * _homeDir )
{
	FILE *fileHandle;
	char workingFile [ MAXCHARS ];
	char textString [ MAXCHARS ];
	char appName [ MAXCHARS ];			// application
	char appResourceName [ MAXCHARS ];	// application's user resource file
	int count;
	unsigned int appCount;
	char c;
	App temp;

	strcpy ( workingFile, _homeDir );
	strcat ( workingFile, "/.blackhome/config/appDefsList" );

	if ( ( fileHandle = fopen ( workingFile, "r" ) ) != NULL )
	{
		c = 0;
		count = 0;
		appCount = 0;
		
		do
		{
			fread ( &c, 1, 1, fileHandle );
			textString [ count ] = c;

			if ( c == 0x0a ) // got current line
			{
				textString [ count ] = 0;

				if ( strlen ( textString ) != 0 )
				{
					if ( textString [ 0 ] != '#' )
					{	// reject comments
						appCount = 0;
						
						while ( textString [ appCount ] != ',' )
							appCount++;	// skip base resource name
							
						appCount++;
						unsigned int ptr = 0;

						while ( appCount < strlen ( textString ) )
						{
							appResourceName [ ptr ] = textString [ appCount ];
							appCount++;
							ptr++;
						}
					
						appResourceName [ ptr ] = 0;
						ptr = 0;

						if ( ( strstr ( appResourceName, "-" ) ) ||
								( strstr ( appResourceName, "." ) ) ||
								 ( strstr ( appResourceName, "_" ) ) )
						{	// app resource has an extension, 'color', probably
							while ( appResourceName [ ptr ] != '-' )
							{
								appName [ ptr ] = appResourceName [ ptr ];
								ptr++;
							}
							
							appName [ ptr ] = 0;
						}
						else
							strncpy ( appName, appResourceName, sizeof ( appName ) );
							
						for ( ptr = 0; ptr < strlen ( appName ); ptr++ )
						{	// might be an internationalisation problem
							c = tolower ( appName [ ptr ] );
							appName [ ptr ] = c;
						}	// attempts to ensure appName is the name of the binary
						
						temp.name = new char [ strlen ( appName ) + 1 ];
						temp.resourceName = new char [ strlen ( appResourceName ) + 1 ];

						strcpy ( temp.name, appName );
						strcpy ( temp.resourceName, appResourceName );
						rfeObjectData.apps.Add ( temp );
					}
				}
				
				count = -1;
			}
			
			count++;
		
		}  while ( !feof ( fileHandle ) );

		fclose ( fileHandle );
	}
	else
	{
		strcpy ( textString, "ResourceFileEditor::listApps\n" );
		strcat ( textString, "Cannot open " );
		strcat ( textString, workingFile );
		strcat ( textString, "\nUnable to construct applications list\n" );
		tablet ( "ResourceFileEditor_error", "file error", textString );
		return 1;
	}

	return 0;
}


int ResourceFileEditor::listAliases ( char * _homeDir, void * resources)
{
	FILE *fileHandle;
	char workingFile [ MAXCHARS ];
	char textString [ MAXCHARS ];
	int count, lineCount;
	char c;
	char * temp;
	rfeResources * myResources = ( rfeResources * ) resources;

	strcpy ( workingFile, _homeDir );
	strcat ( workingFile, "/" );
	strcat ( workingFile, myResources->resourceDir );
	strcat ( workingFile, "/.colourAliasList" );

	if ( ( fileHandle = fopen ( workingFile, "r" ) ) != NULL )
	{
		c = 0;
		count = 0;
		lineCount = 0;
				
		do
		{
			fread ( &c, 1, 1, fileHandle );
			textString [ count ] = c;

			if ( c == 0x0a ) // got current line
			{			
				textString [ count ] = 0;

				if ( strlen ( textString ) != 0 )
				{
					if ( textString [ 0 ] != '!' )
					{	// reject comments
						if ( lineCount == 2 )
							lineCount = -1;

						if ( ( lineCount > 0 ) && ( textString [ 0 ] != 'u' ) && ( textString [ 0 ] != ' ' ) )
						{	// don't list 'u <name>' - is unallocated colour name!
							temp = new char [ strlen ( textString ) + 1 ];
							strcpy ( temp, textString );
							rfeObjectData.aliases.Add ( temp );
						}
						
						lineCount++;
					}
				}
				
				count = -1;
			}
			
			count++;
		
		}  while ( !feof ( fileHandle ) );

		fclose ( fileHandle );
	}
	else
	{
		strcpy ( textString, "ResourceFileEditor::listThemes\n" );
		strcat ( textString, "Cannot read " );
		strcat ( textString, workingFile );
		strcat ( textString, "\nUnable to construct theme list\n" );
		tablet ( "ResourceFileEditor_error", "file error", textString );
		return 1;
	}

	return 0;
}


int ResourceFileEditor::readPrefs ( char * _homeDir )
{
	FILE *stream;
	int error;
	char prefsFile [ MAXCHARS ];
	char header [ MAXCHARS ];
	unsigned int t1, t2, t3, t4, t5, t6, t7;
	float f1, f2, f3, f4, f5, f6, f7;
	char textString [ EDITBUFSIZE ];	// error report
	char numString [ MAXCHARS ];

	strcpy ( prefsFile, _homeDir );
	strcat ( prefsFile, "/.blackhome/config/.resourceFileEditorrc" );

	stream = fopen ( prefsFile, "r" );

	if ( !stream )
	{
		strcpy ( textString, "ResourceFileEditor::readPrefs\n");
		strcat ( textString, "Cannot read ");
		strcat ( textString, prefsFile );
		tablet ( "ResourceFileEditor_error", "file error", textString );
		currentEditLine [ 0 ] = 0;
		return 1;
	}
	
	strcpy ( textString, "read prefs failed.\n");

	error = fscanf ( stream, "%s %u %u", header, &t1, &t2 );

	if ( ( error == 0 ) || ( error == ( int ) EOF ) )
	{
		strcat ( textString, header );
		strcat ( textString, "\n error: " );
		sprintf ( numString, "%d\n", error );
		strcat ( textString, numString );
	}
	else
	{			
		if ( strcmp ( header, "colorSpecSize" ) == 0 )
		{	// allowed deviation from specified colour for search
			_returnSize = t1;
			_fReturnSize = t2;
		}
		else
		{
			tablet ( "ResourceFileEditor_error", "file error",
					"Can't find colorSpecSize in\n.resourceFileEditorrc\n" );
			currentEditLine[ 0 ] = 0;
		}
	}

	error = fscanf ( stream, "%s %f %f %f %f %f %f %f", header,
					&f1, &f2, &f3, &f4, &f5, &f6, &f7 );

	if ( ( error == 0 ) || ( error == ( int ) EOF ) )
	{
		strcat ( textString, header );
		strcat ( textString, "\n error: " );
		sprintf ( numString, "%d\n", error );
		strcat ( textString, numString );
	}
	else
	{			
		if ( strcmp ( header, "colorBounds" ) == 0 )
		{	// allowed deviation from specified colour for search


			if ( f7 < 0 )
				f7 = 0;

			if ( f7 > 1 )
				f7 = 1;		
			
			rfeObjectData.globalColourBound = f7;

			if ( f7 != 0 )
				f1 = f2 = f3 = f4 = f5 = f6 = f7;

			rfeObjectData.redMin = f1;
			rfeObjectData.redMax = f2;
			rfeObjectData.greenMin = f3;
			rfeObjectData.greenMax = f4;
			rfeObjectData.blueMin = f5;
			rfeObjectData.blueMax = f6;
		}
		else
		{
			tablet ( "ResourceFileEditor_error", "file error",
					"Can't find colourBounds in\n.resourceFileEditorrc\n" );
			currentEditLine[ 0 ] = 0;
		}
	}

	error = fscanf ( stream, "%s %u %u %u %u %u %u %u", header,
					&t1, &t2, &t3, &t4, &t5, &t6, &t7);

	if ( ( error == 0 ) || ( error == ( int ) EOF ) )
	{
		strcat ( textString, header );
		strcat ( textString, "\n error: " );
		sprintf ( numString, "%d\n", error );
		strcat ( textString, numString );
	}
	else
	{			
		if ( strcmp ( header, "searchDirs" ) == 0 )
		{	// allowed deviation from specified colour for search
			_searchXdefaults = t1;
			_searchGtkdefaults = t2;
			_searchAppdefaults = t3;
			_searchMenus = t4;
			_searchStyles = t5;
			_appendSearch = t6;
			_keepOldSearchTargets = t7;
		}
		else
		{
			tablet ( "ResourceFileEditor_error", "configuration file error",
						"Can't find searchDirs in\n.resourceFileEditorrc\n" );
			currentEditLine[ 0 ] = 0;
		}
	}

	error = fscanf ( stream, "%s %u", header, &t1 );

	if ( ( error == 0 ) || ( error == ( int ) EOF ) )
	{
		strcat ( textString, header );
		strcat ( textString, "\n error: " );
		sprintf ( numString, "%d\n", error );
		strcat ( textString, numString );
	}
	else
	{			
		if ( strcmp ( header, "defaultColourType" ) == 0 )
		{	// allowed deviation from specified colour for search
			_defaultColourType = t1;
		}
		else
		{
			tablet ( "ResourceFileEditor_error", "configuration file error",
						"Can't find defaultColourType in\n.resourceFileEditorrc\n" );
			currentEditLine[ 0 ] = 0;
		}
	}

	error = fscanf ( stream, "%s %u", header, &t1 );

	if ( ( error == 0 ) || ( error == ( int ) EOF ) )
	{
		strcat ( textString, header );
		strcat ( textString, "\n error: " );
		sprintf ( numString, "%d\n", error );
		strcat ( textString, numString );
	}
	else
	{			
		if ( strcmp ( header, "fallbackColourType" ) == 0 )
		{	// allowed deviation from specified colour for search
			_fallbackColourType = t1;
		}
		else
		{
			tablet ( "ResourceFileEditor_error", "configuration file error",
						"Can't find fallbackColourType in\n.resourceFileEditorrc\n" );
			currentEditLine[ 0 ] = 0;
		}
	}

	error = fscanf ( stream, "%s %u", header, &t1 );

	if ( ( error == 0 ) || ( error == ( int ) EOF ) )
	{
		strcat ( textString, header );
		strcat ( textString, "\n error: " );
		sprintf ( numString, "%d\n", error );
		strcat ( textString, numString );
	}
	else
	{			
		if ( strcmp ( header, "execOnModify" ) == 0 )
		{	// allowed deviation from specified colour for search
			_execOnModify = t1;
		}
		else
		{
			tablet ( "ResourceFileEditor_error", "configuration file error",
						"Can't find execOnModify in\n.resourceFileEditorrc\n" );
			currentEditLine[ 0 ] = 0;
		}
	}

	error = fscanf ( stream, "%s %u %u", header, &t1, &t2 );

	if ( ( error == 0 ) || ( error == ( int ) EOF ) )
	{
		strcat ( textString, header );
		strcat ( textString, "\n error: " );
		sprintf ( numString, "%d\n", error );
		strcat ( textString, numString );
	}
	else
	{			
		if ( strcmp ( header, "shadeReplace" ) == 0 )
		{	// allowed deviation from specified colour for search
			_shadeReplace = t1;
			_generateHue = t2;
		}
		else
		{
			tablet ( "ResourceFileEditor_error", "configuration file error",
						"Can't find shadeReplace in\n.resourceFileEditorrc\n" );
			currentEditLine[ 0 ] = 0;
		}
	}

	error = fscanf ( stream, "%s %f", header, &f1 );

	if ( ( error == 0 ) || ( error == ( int ) EOF ) )
	{
		strcat ( textString, header );
		strcat ( textString, "\n error: " );
		sprintf ( numString, "%d\n", error );
		strcat ( textString, numString );
	}
	else
	{			
		if ( strcmp ( header, "shadeOffset" ) == 0 )
		{	// allowed deviation from specified colour for search
			_shadeOffset = f1;
		}
		else
		{
			tablet ( "ResourceFileEditor_error", "configuration file error",
						"Can't find shadeOffset in\n.resourceFileEditorrc\n" );
			currentEditLine[ 0 ] = 0;
		}
	}

	error = fscanf ( stream, "%s %u", header, &t1 );

	if ( ( error == 0 ) || ( error == ( int ) EOF ) )
	{
		strcat ( textString, header );
		strcat ( textString, "\n error: " );
		sprintf ( numString, "%d\n", error );
		strcat ( textString, numString );
	}
	else
	{			
		if ( strcmp ( header, "saveResourceMods" ) == 0 )
		{	// allowed deviation from specified colour for search
			_saveResourceMods = t1;
		}
		else
		{
			tablet ( "ResourceFileEditor_error", "configuration file error",
						"Can't find saveResourceMods in\n.resourceFileEditorrc\n" );
			currentEditLine[ 0 ] = 0;
		}
	}

	error = fscanf ( stream, "%s %u", header, &t1 );
	

	if ( ( error == 0 ) || ( error == ( int ) EOF ) )
	{
		strcat ( textString, header );
		strcat ( textString, "\n error: " );
		sprintf ( numString, "%d\n", error );
		strcat ( textString, numString );
	}
	else
	{			
		if ( strcmp ( header, "useFirstResource" ) == 0 )
		{	// allowed deviation from specified colour for search
			_useFirstResource = t1;
		}
		else
		{
			tablet ( "ResourceFileEditor_error", "configuration file error",
						"Can't find useFirstResource in\n.resourceFileEditorrc\n" );
			currentEditLine[ 0 ] = 0;
		}
	}

	if ( strcmp ( textString, "read prefs failed.\n" ) != 0 )
	{	// report error(s)
		clearerr ( stream );
		fclose ( stream );

		tablet ( "ResourceFileEditor_error", "file error", textString );
		textString [ 0 ] = 0;
		currentEditLine [ 0 ] = 0;
		return 1;
	}

	if ( ferror ( stream ) )
	{
		fclose ( stream );
		tablet ( "ResourceFileEditor_error", "file error", "ResourceFileEditor::readPrefs stream i/o failed\n" );
		currentEditLine [ 0 ] = 0;
		return 1;
	}
	else
	{
		fclose ( stream );
		return 0;
	}
}


int ResourceFileEditor::writePrefs ( char * _homeDir )
{
	FILE *stream;
	char prefsFile [ MAXCHARS ];
	char textString [ MAXCHARS ];

	strcpy ( prefsFile, _homeDir );
	strcat ( prefsFile, "/.blackhome/config/.resourceFileEditorrc" );

	// initialise file access to filename
	stream = fopen ( prefsFile, "w" );

	if ( !stream )
	{
		strcpy ( textString, "ResourceFileEditor::writePrefs\n");
		strcat ( textString, "Cannot write ");
		strcat ( textString, prefsFile );
		tablet ( "ResourceFileEditor_error", "file error", textString );
		return 1;
	}

	fprintf ( stream, "%s\n%u\n%u\n", "colorSpecSize",
					 	_returnSize, _fReturnSize );

	fprintf ( stream, "%s\n%f\n%f\n%f\n%f\n%f\n%f\n%f\n", "colorBounds",
					 	rfeObjectData.redMin, rfeObjectData.redMax,
						rfeObjectData.greenMin, rfeObjectData.greenMax,
						rfeObjectData.blueMin, rfeObjectData.blueMax,
						rfeObjectData.globalColourBound );

	fprintf ( stream, "%s\n%u\n%u\n%u\n%u\n%u\n%u\n%u\n", "searchDirs",
					 	_searchXdefaults, _searchGtkdefaults, _searchAppdefaults,
						_searchMenus, _searchStyles, _appendSearch,
						_keepOldSearchTargets );
						
	fprintf ( stream, "%s\n%u\n%s\n%u\n", "defaultColourType",
						_defaultColourType, "fallbackColourType",
						_fallbackColourType );

	fprintf ( stream, "%s\n%u\n", "execOnModify", _execOnModify );

	fprintf ( stream, "%s\n%u\n%u\n", "shadeReplace", _shadeReplace, _generateHue );

	fprintf ( stream, "%s\n%f\n", "shadeOffset", _shadeOffset );

	fprintf ( stream, "%s\n%u\n", "saveResourceMods", _saveResourceMods );

	fprintf ( stream, "%s\n%u\n", "useFirstResource", _useFirstResource );

	if ( ferror ( stream ) )
	{
		clearerr ( stream );
		fclose ( stream );
		strcpy ( textString, "ResourceFileEditor::writePrefs\n" );
		strcat ( textString, "write failed.\n" );
		tablet ( "ResourceFileEditor_error", "file error", textString );
		currentEditLine [ 0 ] = 0;
		return 1;
	}
	else
	{
		fclose ( stream );
		return 0;
	}
}


char * ResourceFileEditor::tablet ( char * title, char *labelText, char * text )
{	// if supplied with text, display and return nothing, else return text input
	Widget _textPopup;
	Arg args [ 2 ];   

	// shell resources lost in another thread...
	XtSetArg ( args [ 0 ], XtNwidth, BOXWIDTH );
	XtSetArg ( args [ 1 ], XtNheight, BOXHEIGHT );

	_textPopup = XtCreateManagedWidget ( title, shellWidgetClass,
											_popup, args, 2 );
	Text myText ( title, _textPopup );	

	if ( strlen ( text ) > 0 )
		myText.setText ( text );

	myText.setLabelText ( labelText );
	
	XtPopupSpringLoaded ( ( Widget ) myText._popup ); // popup widget
	XEvent myTextEvent; // wait on Text widget event

	while ( myText.getState() == TEXT_NOT_USEFUL )
	{
		XtAppNextEvent ( _app, &myTextEvent );
		XtDispatchEvent ( &myTextEvent );
	}
   
	XtPopdown ( ( Widget ) myText._popup );

	if ( myText.getState() == TEXT_CANCEL_PRESSED )
	{
		myText.setStateToNOTUseful();		
		strcpy ( currentEditLine, "tabletCancel" );
	}		
   
	if ( myText.getState() == TEXT_OK_PRESSED )
	{
		if ( strlen ( myText.getText() ) < EDITBUFSIZE )
			strcpy ( currentEditLine, myText.getText() );

		myText.setStateToNOTUseful();
	}

	return ( char * ) currentEditLine;
}


int ResourceFileEditor::menu ( char *title, char * popupTitle, char * menuTitle, char * menuItems )
{	// menuItems: comma separated format, trailing comma
	// return codes: -1 error, 1,... menu item number selected
	Widget _menuPopup;
	char textString [ MAXCHARS ];
	unsigned int numItems = 0;
	unsigned int maxItemLength = 0;
	unsigned int posn;
	unsigned int stringLength;
	unsigned int menuWidth; // kludge for shell resources lost in another thread
	unsigned int menuHeight;
	int count;
	Arg args [ 2 ];

	menuWidth = 10 * strlen ( popupTitle );
	menuHeight = ( unsigned int ) ( menuWidth + 0.5 ) / 4;
	
	XtSetArg ( args [ 0 ], XtNwidth, menuWidth );
	XtSetArg ( args [ 1 ], XtNheight, menuHeight );

	stringLength = strlen ( menuItems );

	// find number of items and max item length
	count = 0;
	posn = 0;
	
	while ( posn < stringLength )
	{
		textString[count] = menuItems [ posn ];
			
		if (textString[count] == ',') // found item separator
		{
			numItems++; // got one
			textString [ count ] = 0;

			if (strlen(textString) > maxItemLength)
				maxItemLength = strlen(textString);

			count = -1;
		}

		count++;
		posn++;
	}

	// set up string array to hold menu item list
	char strings [ numItems ] [ maxItemLength + 1 ];

	count = 0;
	posn = 0;
	int itemCount = 0;
		
	while ( posn < stringLength )
	{
		textString[count] = menuItems [ posn];

		if (textString[count] == ',')
		{
			textString[count] = 0;

			if (strlen(textString) != 0)
			{
				strcpy ( strings[itemCount], textString);
				itemCount++;
 				count = -1;
			}
		}
			
		count++;
		posn++;
	}

	_menuPopup = XtCreateManagedWidget ( title, topLevelShellWidgetClass,
											_popup, args, 2 );

	Menu myMenu ( "myMenu", _menuPopup, popupTitle, menuTitle,
					numItems, maxItemLength+1, ( char * ) &strings );	

	XtPopup ( ( Widget ) myMenu._popup, XtGrabNone ); // popup widget
	XEvent myMenuEvent;	// wait on menu widget event

	while (myMenu.getState() == MENU_NOT_USEFUL )
	{
		XtAppNextEvent ( _app, &myMenuEvent );
		XtDispatchEvent ( &myMenuEvent );
	}

	XtPopdown ( ( Widget ) myMenu._popup );

	if ( myMenu.getState() == MENU_CANCEL_PRESSED )
		myMenu.setStateToNOTUseful();

	if ( myMenu.getState() == MENU_OK_PRESSED )
	{
		myMenu.setStateToNOTUseful();
		_menuItem =  myMenu.getSelectedItem();
		return _menuItem;
	}

	return 0;	// menu items start at 1...
}


int ResourceFileEditor::reportColour ( ColourSpecConvert * myConvert )
{
	char textString [ MAXCHARS ];
	char numString [ MAXCHARS ];
		
	strcpy ( textString, "integer:     " );
	sprintf ( numString, "%s", myConvert->intColour() );
	strcat ( textString, numString );

	strcat ( textString, "\ntruncated:   " );
	sprintf ( numString, "%s", myConvert->hashColour() );
	strcat ( textString, numString );

	strcat ( textString, "\nrgb:         " );
	sprintf ( numString, "%s", myConvert->rgbColour() );
	strcat ( textString, numString );

	strcat ( textString, "\nrgbi:        " );
	sprintf ( numString, "%s", myConvert->rgbiColour() );
	strcat ( textString, numString );

	strcat ( textString, "\nnamed:       " );
	sprintf ( numString, "%s", myConvert->namedColour() );
	strcat ( textString, numString );

	strcat ( textString, "\nanalysis:    " );
	sprintf ( numString, "%s", myConvert->analyseColour() );
	strcat ( textString, numString );

	strcat ( textString, "\nalias:       " );
	sprintf ( numString, "%s", myConvert->aliasColour() );
	strcat ( textString, numString );

	strcat ( textString, "\0" );
	tablet ( "ColourSpecConvert_report", "colour equivalents", textString );
	
	return 0;
}


void ResourceFileEditor::reportColourError ( char * colourString, int error )
{
	char textString [ EDITBUFSIZE ];
	char numString [ MAXCHARS ];
	
	strcpy ( textString, "Error " );
	sprintf ( numString, "%d", error );
	strcat ( textString, numString );
	strcat ( textString, " in colour specification " );
	sprintf ( numString, "%s:\n", colourString );
	strcat ( textString, numString );

	switch ( error )
	{
		case ERR_STRINGTOOLONG:
			strcat ( textString, "String length exceeds " );
			sprintf( numString, "%d\n", MAXCHARS );
			strcat ( textString, numString );
			break;
			
		case ERR_COLOURTYPE:
			strcat ( textString, "Colour type unrecognised.\n" );
			break;
			
		case ERR_DEC:
			strcat ( textString, "decimal number error.\n" );
			break;

		case ERR_HEX:
			strcat ( textString, "hexadecimal number error.\n" );
			break;

		case ERR_LETTER:
			strcat ( textString, "expecting something recognisably\nin the English alphabet.\n" );
						
		case ERR_SEPARATOR:
			strcat ( textString, "missing or invalid separator.\n" );
			break;
			
		case ERR_DP:
			strcat ( textString, "surplus decimal points, confused...\n" );
			break;
			
		case ERR_MISSING_DIGIT:
			strcat ( textString, "missing digit.\n" );
			break;
			
		case ERR_RGBTXT:
			strcat ( textString, "rgb.txt: path or file name error.\n" );
			break;
			
		case ERR_FILENOTFOUND:
			strcat ( textString, "path error; cannot locate data file.\n" );
			break;
			
		case ERR_COLOURSEP:
			strcat ( textString, "ColourSpecConvert::separateColour failed.\n");

		case ERR_RESPATH:
			strcat ( textString, "resource file path not specified.\n");

		case ERR_COLOURDEFSLIST:
			strcat ( textString, "can't open .colourAliasList file.\n");
			
		default:
			strcat ( textString, "bug: error in programmer\nemail will@opaline.demon.co.uk\n" );
	}

	tablet ( "ColourSpecConvert_error", "colour specification error", textString );
}


void ResourceFileEditor::cleanOperate ( void )
{
	ListItem < Ops > *opItem = rfeObjectData.opOn.GetHead();
	
	while ( opItem )
	{
		delete[] opItem->Get().fileName;
		delete[] opItem->Get().resourceLabel;
		delete[] opItem->Get().returnedResource;
		delete[] opItem->Get().specifiedResource;
		opItem = opItem->Next();
	}

	rfeObjectData.opOn.Release();

}


void ResourceFileEditor::cleanSearchInFiles ( void )
{
	ListItem < char * > *fileItem = rfeObjectData.targets.GetHead();
	
	while ( fileItem )
	{
		delete[] fileItem->Get();
		fileItem = fileItem->Next();
	}

	rfeObjectData.targets.Release();
}


void ResourceFileEditor::cleanSelApps ( void )
{
	ListItem < char * > *selAppItem = rfeObjectData.selApps.GetHead();
	
	while ( selAppItem )
	{
		delete[] selAppItem->Get();
		selAppItem = selAppItem->Next();
	}

	rfeObjectData.selApps.Release();
}


void ResourceFileEditor::cleanSelThemes ( void )
{
	ListItem < char * > *selThemeItem = rfeObjectData.selThemes.GetHead();

	selThemeItem = rfeObjectData.selThemes.GetHead();

	while ( selThemeItem )
	{
		delete[] selThemeItem->Get();
		selThemeItem = selThemeItem->Next();
	}

	rfeObjectData.selThemes.Release();
}


void ResourceFileEditor::cleanAvailable ( void )
{	// should there be listItems for ,name and .resourceName separately?
	ListItem < char * > *themeItem = rfeObjectData.themes.GetHead();

	while ( themeItem )
	{
		delete[] themeItem->Get();
		themeItem = themeItem->Next();
	}

	rfeObjectData.themes.Release();

	ListItem < App > *appItem = rfeObjectData.apps.GetHead();

	while ( appItem )
	{
		delete[] appItem->Get().name;
		delete[] appItem->Get().resourceName;
		appItem = appItem->Next();
	}
	
	rfeObjectData.apps.Release();

	ListItem < char * > *aliasItem = rfeObjectData.aliases.GetHead();

	while ( aliasItem )
	{
		delete[] aliasItem->Get();
		aliasItem = aliasItem->Next();
	}

	rfeObjectData.aliases.Release();
}


void ResourceFileEditor::cleanReplace ( void )
{
	ListItem < char * > *replaceItem = rfeObjectData.replace.GetHead();
	
	while ( replaceItem )
	{
		delete[] replaceItem->Get();
		replaceItem = replaceItem->Next();
	}

	rfeObjectData.replace.Release();
}


void ResourceFileEditor::memError ( int availableSize, char * where )
{
	sprintf( currentEditLine, "%d", availableSize );
	strcat ( currentEditLine, " byte buffer size exceeded in\n" );
	strcat ( currentEditLine, where );
	strcat ( currentEditLine, ".\nChange this in 'Global.h' and recompile\n" );
	strcat ( currentEditLine, "resourceFileEditor.\n" );
	tablet ( "ResourceFileEditor_error", "potential buffer overflow!", currentEditLine );
	currentEditLine [ 0 ] = 0;
}


char * ResourceFileEditor::shadeColour ( char * specifiedColour, char * newColour )
{
	float shift;
	float oldIntensity, newIntensity;
	float sRed, sGreen, sBlue, nRed, nGreen, nBlue;
	float red, green, blue;
	int error;
	char textString [ MAXCHARS ];
	char numString [ MAXCHARS ];
	
	ColourSpecConvert *myConvert = new ColourSpecConvert;

	myConvert->setRgbTxtPath ( myResources->rgbTxtDir );
	myConvert->setfReturnSize ( _fReturnSize );
	myConvert->setReturnSize ( 16 );

	strcpy ( textString, _homeDir );
	strcat ( textString, "/" );
	strcat ( textString, myResources->resourceDir );

	myConvert->setResourceFilePath ( textString );

	error = myConvert->setColour( specifiedColour );

	if ( error ) 
	{
		reportColourError ( specifiedColour, error );
		currentEditLine [ 0 ] = 0;
		delete myConvert;
		return specifiedColour;
	}

	sRed = myConvert->redF();
	sGreen = myConvert->greenF();
	sBlue = myConvert->blueF();
	
	sRed += 1e-6;
	sGreen += 1e-6;
	sBlue += 1e-6;

	oldIntensity = ( sRed + sGreen + sBlue) / 3;

	error = myConvert->setColour ( newColour );

	if ( error ) 
	{
		reportColourError ( newColour, error );
		currentEditLine [ 0 ] = 0;
		delete myConvert;
		return newColour;	// instead of some garbage, return the specified colour
	}

	nRed = myConvert->redF();
	nGreen = myConvert->greenF();
	nBlue = myConvert->blueF();

	nRed += 1e-6;
	nGreen += 1e-6;
	nBlue += 1e-6;

	newIntensity = ( nRed + nGreen + nBlue) / 3;
	newIntensity = newIntensity / oldIntensity;
	
	if ( newIntensity > 1 )
		newIntensity = 1 / newIntensity;

	if ( _generateHue )
	{
		if ( ( nRed >= nGreen ) && ( nRed >= nBlue ) )
		{	// red dominant
			red = nRed * newIntensity;
			green = nGreen * ( ( sGreen / sRed ) * red );	// preserve dominance
			blue = nBlue * ( ( sBlue / sRed ) * red );		// change hue slightly
		}
		else
		{
			if ( ( nGreen >= nRed ) && ( nGreen >= nBlue ) )
			{	//green dominant
				green = nGreen * newIntensity;
				red = nRed * ( ( sRed / sGreen ) * green );
				blue = nBlue * ( ( sBlue / sGreen ) * green );
			}
			else
			{	// blue dominant
				blue = nBlue * newIntensity;
				red = nRed * ( ( sRed / sBlue ) * blue );
				green = nGreen * ( ( sGreen / sBlue ) * blue );
			}
		}
	}
	else
	{	// just vary shade
		if ( ( nRed >= nGreen ) && ( nRed >= nBlue ) )
		{	// red dominant
			red = nRed  * newIntensity;
			green = nGreen * newIntensity;	// preserve dominance
			blue = nBlue * newIntensity;	// preserve hue
		}
		else
		{
			if ( ( nGreen >= nRed ) && ( nGreen >= nBlue ) )
			{	//green dominant
				green = nGreen * newIntensity;
				red = nRed * newIntensity;
				blue = nBlue * newIntensity;
			}
			else
			{	// blue dominant
				blue = nBlue * newIntensity;
				red = nRed * newIntensity;
				green = nGreen * newIntensity;
			}
		}
	}

	red += _shadeOffset;
	green += _shadeOffset;
	blue += _shadeOffset;

	shift = 0;
	
	if ( red < shift )
		shift = red;

	if ( green < shift )
		shift = green;
		
	if ( blue < shift )
		shift = blue;
		
		red -= shift;
		green -= shift;
		blue -= shift;
	
	shift = 1;
	
	if ( red > shift )
		shift = red;
		
	if ( green > shift )
		shift = green;
		
	if ( blue > shift )
		shift = blue;

	red /= shift;
	green /= shift;
	blue /= shift;

	sprintf ( numString, "%f/%f/%f\n", red, green, blue );
	strcpy ( textString, "rgbi:" );
	strcat ( textString, numString );

	error = myConvert->setColour ( textString );
	
	if ( error ) 
		strcpy ( currentEditLine, newColour );
	else
		strcpy ( currentEditLine, textString );

	delete myConvert;
	return ( char * ) currentEditLine;	// return with new colour value set in ColourSpecConvert	
}


char * ResourceFileEditor::formatColour ( char *oldColour )
{
	char textString [ MAXCHARS ];
	int error;
	unsigned int p1, p2;
	
	ColourSpecConvert *myConvert = new ColourSpecConvert;
	
	myConvert->setRgbTxtPath ( myResources->rgbTxtDir );
	myConvert->setfReturnSize ( _fReturnSize );
	myConvert->setReturnSize ( _returnSize );

	strcpy ( textString, _homeDir );
	strcat ( textString, "/" );
	strcat ( textString, myResources->resourceDir );

	myConvert->setResourceFilePath ( textString );

	error = myConvert->setColour ( oldColour );

	if ( error ) 
	{
		reportColourError ( oldColour, error );
		currentEditLine [ 0 ] = 0;
		delete myConvert;
		return "";
	}

	// now substitute resource for option-determined type
	switch ( _defaultColourType )
	{	// this also corrects for excess spaces in input
		case 0:
			strcpy ( currentEditLine, myConvert->hashColour() );
			break;

		case 1:
			strcpy ( currentEditLine, myConvert->rgbColour() );
			break;

		case 2:
			strcpy ( currentEditLine, myConvert->rgbiColour() );
			break;

		case 3:
			{	// using 'named' as default
				if ( strlen ( myConvert->namedColour() ) == 0 )
				{	// named resource fails, get fallback
					switch ( _fallbackColourType )
					{
						case 0:
							strcpy ( currentEditLine, myConvert->hashColour() );
							break;

						case 1:
							strcpy ( currentEditLine, myConvert->rgbColour() );
							break;

						case 2:
							strcpy ( currentEditLine, myConvert->rgbiColour() );
							break;
					}
				}
				else
				{
					strcpy ( textString, myConvert->namedColour() );
					// strip spaces, avoids complications testing later
					p1 = p2 = 0;
					
					while ( p1 < strlen ( textString ) )
					{
						if ( textString [ p1 ] != ' ' )
						{
							currentEditLine [ p2 ] = textString [ p1 ];
							p1++;
							p2++;
						}
						else
							p1++;
					}
					
					currentEditLine [ p2 ] = 0;
				}
			}

			break;
	}	// end switch _defaultColourType

	if ( error )
		currentEditLine [ 0 ] = 0;

	delete myConvert;
	return ( char * ) currentEditLine;
}


int ResourceFileEditor::substituteResource ( void )
{
	FILE *fileHandle;
	char * resourceFile;
	char textString [ EDITBUFSIZE ];
	char temp [ MAXCHARS ];
	char workingFile [ MAXCHARS ];			
	char resource [ MAXCHARS ];
	char returnedResource [ MAXCHARS ];
	char c, checkChar;
	long ptr, curPos;
	int error;
	long fileCount;
	long srcLength, dstLength, count;
	long posInFile, diff;
	unsigned int lineCount;
	bool isGnomeColour;
	char gnomeColour [ MAXCHARS ];
	int fileType;	// 0 std colour, 1 gnome spec '{ r, g, b }'
	
	snprintf ( _gnomeColourFormatString, sizeof ( _gnomeColourFormatString ), "%s%d%s%d%s%d%s", "{ %.",
		_fReturnSize, "f, %.", _fReturnSize, "f, %.", _fReturnSize, "f }" );

	for ( fileCount = 0; fileCount < rfeObjectData.opOn.Count(); fileCount++ )
	{
		strncpy ( returnedResource, rfeObjectData.opOn[ fileCount ].returnedResource, sizeof ( returnedResource ) );
		strcpy ( workingFile, rfeObjectData.opOn[ fileCount ].fileName );

		if ( ( fileHandle = fopen ( workingFile, "r" ) ) == NULL )
		{
			strcpy ( textString, "File open failed: " );
			strcat ( textString, workingFile );
			tablet ( "ResourceFileEditor_error", "file error", textString );
			currentEditLine [ 0 ] = 0;
			textString [ 0 ] = 0;
			return 1;
		}

		if ( strstr ( workingFile, "gtkrc-" ) != NULL )
			fileType = 1;
		else
			fileType = 0;

		dstLength = 0;
		fseek ( fileHandle, 0L, SEEK_END);
		srcLength = ftell ( fileHandle );
		fseek ( fileHandle, 0L, SEEK_SET );
		c = 0;		
		count = 0;
		posInFile = 0;

		// find the length of the new resource file to be written
		while ( !feof ( fileHandle ) )
		{
			fread ( &c, 1, 1, fileHandle );
			textString [ count ] = c;

			if ( c == 0x0a )	// got current line	
			{
				isGnomeColour = FALSE;
				textString [ count + 1 ] = 0;

				switch ( fileType )
				{
					case 1:		// gnome colour
					char junk [ MAXCHARS ];
					float gtkR,gtkG,gtkB;
					float r, g, b;
					// The undo file can be written with returned resource in non-rgbi
					// format (default is rgb). So the test fails if the colour isn't converted.
					if ( strncmp ( returnedResource, "rgbi", 4 ) != 0  )
					{	// undo file colours written in rfe colour format (default rgb)
						ColourSpecConvert *myConvert = new ColourSpecConvert;

						myConvert->setfReturnSize ( _fReturnSize );
						myConvert->setReturnSize ( _returnSize );

						error = myConvert->setRgbTxtPath ( myResources->rgbTxtDir );

						if ( error )
							reportColourError ( myResources->rgbTxtDir, error );

						myConvert->setColour ( returnedResource );
						strncpy ( returnedResource, myConvert->rgbiColour(), sizeof ( returnedResource ) );
						delete myConvert;
					}

					sscanf ( returnedResource, "rgbi:%f/%f/%f", &r, &g, &b );

					if ( sscanf ( textString, "%s = { %f, %f, %f }",&junk, &gtkR, &gtkG, &gtkB ) == 4 )
					{
						if ( ( r == gtkR ) && ( g == gtkG ) && ( b == gtkB ) )
						{
							isGnomeColour = TRUE;
							curPos = strstr ( textString, "] = {" ) - textString + 4;	// start of colour spec

							int length = strlen ( textString ) - curPos - 1;

							for ( ptr = 0; ptr < length; ptr++ )
								resource [ ptr ] = textString [ ptr + curPos ];

							resource [ length ] = 0;
							diff = strlen ( textString ) - strlen ( resource );
							diff += strlen ( gnomeColour );
							dstLength += diff;
						}
					}	// end extract gnome colour

					if  (isGnomeColour == FALSE )
						dstLength += strlen ( textString );
					break;	

					default:	// std colour ( #, rgb, rgbi, named)
					if ( strstr ( textString, returnedResource ) != NULL )
					{	// possibly found resource to replace
						error = 1;	// clear if resource confirmed
						ptr = strstr ( textString, returnedResource ) - textString;
						curPos = 0;

						while ( ptr < ( long ) strlen ( textString ) )
						{
							if ( ( textString [ ptr ] == ' ' ) || ( textString [ ptr ] == '\t' ) || ( textString [ ptr ] == '\n' ) )
								ptr = strlen ( textString );
							else
							{
								resource [ curPos ] = textString [ ptr ];
								ptr++;
								curPos++;
							}
						}

						resource [ curPos ] = 0;

						if ( strcmp ( returnedResource, resource ) == 0 )
							error = 0;	// exact match confirmed to list item

						if ( ! error )
						{
							diff = strlen(textString);
							diff -= strlen ( resource );
							diff += strlen ( rfeObjectData.replace[ fileCount ]);
							dstLength += diff;
						}
					}
					else
						dstLength += strlen ( textString);
				}	// end switch ( fileType )

				count = -1;	// reset for new textString
				textString [ 0 ] = 0;
				c = 0;
			}	// end got current line						

			count++;
			posInFile++;
		}	// end of file, done

		resourceFile = new char [ dstLength<<1 ];	// 2x safety margin...
		resourceFile [ 0 ] = 0;

		fseek ( fileHandle, 0L, SEEK_SET );
		c = 0;		
		count = 0;
		textString [ 0 ] = 0;

		while ( !feof ( fileHandle ) )
		{
			fread ( &c, 1, 1, fileHandle );
			textString [ count ] = c;

			if ( c == 0x0a )	// got current line			
			{
				textString [ count + 1 ] = 0;

				if ( ( textString [ 0 ] == '#' ) || ( textString [ 0 ] == '!' ) )
 					strcat ( resourceFile, textString );
				else
				{
					error = 1;

					if ( ( int ) rfeObjectData.opOn[ fileCount ].resourceType == 2 )
					{	// if this is a colour resource
						if (strstr ( workingFile, "gtkrc-" ) != NULL )
						{
							char junk [ MAXCHARS ];
							float gtkR,gtkG,gtkB;
							float r, g, b;

							isGnomeColour = FALSE;
							sscanf ( returnedResource, "rgbi:%f/%f/%f", &r, &g, &b );

							if ( sscanf ( textString, "%s = { %f, %f, %f }",&junk, &gtkR, &gtkG, &gtkB ) == 4 )
								if ( ( r == gtkR ) && ( g == gtkG ) && ( b == gtkB ) )
									isGnomeColour = TRUE;
						}

						if ( isGnomeColour == FALSE )
						{
							if ( strstr ( textString, rfeObjectData.opOn[ fileCount].returnedResource ) != NULL )
							{
								ptr = strstr ( textString, returnedResource ) - textString;
								curPos = 0;

								while ( ptr < ( long ) strlen ( textString ) )
								{
									if ( ( textString [ ptr ] == ' ' ) || ( textString [ ptr ] == '\t' ) || ( textString [ ptr ] == '\n' ) )
										ptr = strlen ( textString );
									else
									{
										resource [ curPos ] = textString [ ptr ];
										ptr++;
										curPos++;
									}
								}

								resource [ curPos ] = 0;

								if ( strcmp ( returnedResource, resource ) == 0 )
									error = 0;	// exact match confirmed to list item					

								if ( !error )
								{
									strcpy ( temp, resource );
									diff = strstr ( textString, temp ) - textString;
									diff += strlen ( temp );
									diff -= strlen ( textString );
								}					
							}
							else
								error = 1;
						}
						else
						{
							strcpy ( temp, resource );
							error = 0;	// no gnome colour error
						}
					}

					if ( _searchType == 1 )
					{
						if ( strstr ( textString, rfeObjectData.opOn[ fileCount ].resourceLabel ) == NULL )
							error = 1; // found a colour but the labels don't match
					}

					if ( !error )
					{	// found resource to replace
						diff = strstr ( textString, temp ) - textString;
						temp [ 0 ] = 0;
						strncat ( temp, textString, diff );
						diff = strlen ( textString ) - ( diff + strlen ( resource ) );

						if ( isGnomeColour )
						{
							float r, g, b;
							char formatString [ MAXCHARS ];

							ColourSpecConvert *myConvert = new ColourSpecConvert;
							myConvert->setfReturnSize ( _fReturnSize );
							myConvert->setReturnSize ( _returnSize );
	
							error = myConvert->setRgbTxtPath ( myResources->rgbTxtDir );

							if ( error )
							{
								reportColourError ( myResources->rgbTxtDir, error );
								delete myConvert;
								return ( 1 );
							}
	
							myConvert->setColour ( rfeObjectData.replace [ fileCount ] );
							sscanf ( myConvert->rgbiColour(), "rgbi:%f/%f/%f", &r, &g, &b );
							snprintf ( gnomeColour, sizeof ( gnomeColour ), _gnomeColourFormatString, r, g, b );
							strcat ( temp, gnomeColour );
							delete myConvert;
						}
						else
							strcat ( temp, rfeObjectData.replace [ fileCount ] );

						if ( strstr ( workingFile, ".menu" ) != NULL )
							strcat ( temp, " " );

						posInFile = strlen ( temp);
						lineCount = strlen ( textString ) - diff;

						if ( strstr ( workingFile, ".menu" ) != NULL )
						{
							if ( textString [ lineCount ] == ' ' )
								lineCount++;
						}

						while ( lineCount < strlen ( textString ) )
						{
							temp [ posInFile ] = textString [ lineCount ];
							lineCount++;
							posInFile++;
						}

						temp [ posInFile ] = 0;
						strcat ( resourceFile, temp );
					}
					else
						strcat ( resourceFile, textString );

				}	// end if ( !(comment or #include) )

				count = -1;	// reset for new textString
				textString [ 0 ] = 0;
				c = 0;
				checkChar = 0;							
			}	// end got current line

			count++;
		}	// end of file, done

		fclose ( fileHandle );
		resourceFile [ strlen ( resourceFile ) + 1 ] = 0;

		if ( ( fileHandle = fopen ( workingFile, "w" ) ) != NULL )
		{
			fwrite ( resourceFile, 1, strlen ( resourceFile ), fileHandle );
			fflush ( fileHandle );
			fclose ( fileHandle );
		}
		else
		{
			strcpy ( textString, "Cannot write resource file\n" );
			strcat ( textString, workingFile );
			tablet ( "ResourceFileEditor_error", "resource write error", textString );
			currentEditLine [ 0 ] = 0;
			delete[] resourceFile;
			return 1;
		}

		delete[] resourceFile;
	}	// done all targets

	return 0;
}


int ResourceFileEditor::execOnModify ( void )
{
	char workingFile [ MAXCHARS ];
	char textString [ MAXCHARS ];		
	char temp [ MAXCHARS ];
	int count;
	int checkAlias = 0;

	FILE *fileHandle;

	if ( ( _execOnModify ) && ( rfeObjectData.selApps.Count() != 0 ) )
	{
		strcpy ( workingFile, _homeDir );
		strcat ( workingFile, "/" );
		strcat ( workingFile, myResources->resourceDir );
		strcat ( workingFile, "/.appNameAlias" );

		if ( ( fileHandle = fopen ( workingFile, "r" ) ) == NULL )
		{ 
			strcpy ( textString, "cannot open " );
			strcat ( textString, workingFile );
			strcat ( textString, "\nPlease create an empty .appNameAlias file\nas a default at least.\n" );
			tablet ( "ResourceFileEfditor_error", "file error", textString );
			currentEditLine [ 0 ] = 0;
			return 1;
		}
		
		fseek ( fileHandle, 0L, SEEK_END );

		if ( ftell ( fileHandle ) > 0 )
			checkAlias = 1;
		
		fseek ( fileHandle, 0L, SEEK_SET );

		char appName [ MAXCHARS ];
		char aliasName [ MAXCHARS ];
		int appCount;
		char oldApp [ MAXCHARS ];

		for ( count = 0; count < rfeObjectData.selApps.Count(); count++ )
		{
			for ( appCount = 0; appCount < rfeObjectData.apps.Count(); appCount++ )
			{
				if ( strstr ( rfeObjectData.selApps[ count ], rfeObjectData.apps [ appCount ].resourceName ) != NULL )
				{	// get the app name
					if ( strcmp ( rfeObjectData.apps[appCount].name, oldApp ) != 0 )
					{	// only fire up app once for each mention in opOn
						// check the name alias file for a translation
						if ( checkAlias )	// always true, woops
						{	//scan file for match to apps[].name
							fseek ( fileHandle, 0L, SEEK_SET );

							while ( !feof ( fileHandle ) )
							{
								fscanf ( fileHandle, "%s %s", appName, aliasName );

								if ( strcmp ( rfeObjectData.apps[ appCount ].name, appName ) == 0 )
									break;
								else
									aliasName [ 0 ] = 0;
							}
						}

						strcpy ( oldApp, rfeObjectData.apps[ appCount ].name );

						strcpy ( textString, "exec " );

						if ( strlen ( aliasName ) == 0 )
							strcat ( textString, rfeObjectData.apps[ appCount ].name );
						else
						{
							unsigned int src = 0;
							unsigned int dst = 0;

							temp [ 0 ] = 0;

							while ( src < strlen ( aliasName ) )
							{
								if ( aliasName [ src ] == ',' )
									temp [ dst ] = ' ';
								else
									temp [ dst ] = aliasName [ src ];

								src++;
								dst++;
							}

							temp [ dst ] = 0;
							strcat(textString, temp);
						}

						strcat ( textString, " & ");
						sysCall ( textString );
					}
				}
			}
		}

		fclose ( fileHandle );
	}	// end execOnModify

	return 0;
}


unsigned int ResourceFileEditor::addColourTarget ( char * resource, ColourSpecConvert *myConvert,
											unsigned int aliasFlag, char * fileName,
											char * resourceLabel, char * returnedResource,
											float redLo, float redHi,
											float greenLo, float greenHi,
											float blueLo, float blueHi )
{
// 'resource' is the sought-for item; 'resourceLabel' is the widget or app
// resource using 'resource'; returnedResource is a valid resource found in one
// line of the current file, to which 'resource' is compared; successful match
// adds 'returnedResource' as a new target specified by 'resourceLabel'
// return '1' target added, '0' failed
	int error;	
	Ops tempOps;
	
	if ( aliasFlag )
	{
		error = 1; // false, no match

		if ( strcmp ( resource, myConvert->aliasColour() ) == 0 )
			error = 0;
		else
		{
			char madeResource [ MAXCHARS ];

			strcpy ( madeResource, "bright " );
			strcat ( madeResource, resource );

			if ( strcmp ( madeResource, myConvert->aliasColour() ) == 0 )
				error = 0;

			if ( error )
			{
				strcpy ( madeResource, "pale " );
				strcat ( madeResource, resource );

				if ( strcmp ( madeResource, myConvert->aliasColour() ) == 0 )
					error = 0;
			}										

			if ( error )
			{
				strcpy ( madeResource, "light " );
				strcat ( madeResource, resource );

				if ( strcmp ( madeResource, myConvert->aliasColour() ) == 0 )
					error = 0;
			}										

			if ( error )
			{
				strcpy ( madeResource, "mid " );
				strcat ( madeResource, resource );

				if ( strcmp ( madeResource, myConvert->aliasColour() ) == 0 )
					error = 0;
			}

			if ( error )
			{
				strcpy ( madeResource, "deep " );
				strcat ( madeResource, resource );

				if ( strcmp ( madeResource, myConvert->aliasColour() ) == 0 )
					error = 0;
			}

			if ( error )
			{
				strcpy ( madeResource, "dark " );
				strcat ( madeResource, resource );

				if ( strcmp ( madeResource, myConvert->aliasColour() ) == 0 )
					error = 0;
			}										

			if ( error )
			{
				strcpy ( madeResource, "shadow " );
				strcat ( madeResource, resource );

				if ( strcmp ( madeResource, myConvert->aliasColour() ) == 0 )
					error = 0;
			}										
		}

		if ( !error )
		{	// 'is alias'
			tempOps.fileName = new char [ strlen ( fileName ) + 1 ];
			tempOps.resourceLabel = new char [ strlen ( resourceLabel ) + 1 ];
			tempOps.returnedResource = new char [ strlen ( returnedResource ) + 1 ];
			tempOps.specifiedResource = new char [ strlen ( resource ) + 1 ];
			tempOps.resourceType = new unsigned int;

			strcpy ( tempOps.fileName, fileName );
			strcpy ( tempOps.resourceLabel, resourceLabel );
			strcpy ( tempOps.returnedResource, returnedResource );
			strcpy ( tempOps.specifiedResource, resource );
			tempOps.resourceType = (unsigned int * ) 2; // signifies 'colour'
			rfeObjectData.opOn.Add ( tempOps );
		}
		else
			error = 1; // unsuccessful alias match
	}
	else
	{	// 'is almost or equal'
		char string [ MAXCHARS ];
		float sRed, sGreen, sBlue;
		float rRed, rGreen, rBlue;

		// get resource and replacement in float format
		myConvert->setColour ( resource );
		strcpy ( string, myConvert->rgbiColour() );
		sscanf ( string, "rgbi:%f/%f/%f", &sRed, &sGreen, &sBlue );

		myConvert->setColour ( returnedResource );
		strcpy ( string, myConvert->rgbiColour() );
		sscanf ( string, "rgbi:%f/%f/%f", &rRed, &rGreen, &rBlue );

		error = 0;
/*
printf("resource %s returned %s\n", resource, returnedResource);
printf("sRed  sGreen  sBlue  %f %f %f\n", sRed,sGreen,sBlue);
printf("rRed  rGreen  rBlue  %f %f %f\n", rRed,rGreen,rBlue);
printf("redLo greenLo blueLo %f %f %f\n", redLo,greenLo,blueLo);
printf("redHi greenHi blueHi %f %f %f\n", redHi,greenHi,blueHi);
*/
		if ( ( sRed < rRed-redLo ) || ( sRed > rRed+redHi ) )
			error = 1;
	
		if ( ( sGreen < rGreen-greenLo ) || ( sGreen > rGreen+greenHi ) )
			error = 1;

		if ( ( sBlue < rBlue-blueLo ) || ( sBlue > rBlue+blueHi ) )
			error = 1;

//printf("error %d\n\n", error);

		if ( !error )
		{
			tempOps.fileName = new char [ strlen ( fileName ) + 1 ];
			tempOps.resourceLabel = new char [ strlen ( resourceLabel ) + 1 ];
			tempOps.returnedResource = new char [ strlen ( returnedResource ) + 1 ];
			tempOps.specifiedResource = new char [ strlen ( resource ) + 1 ];
			tempOps.resourceType = new unsigned int;

			strcpy ( tempOps.fileName, fileName );
			strcpy ( tempOps.resourceLabel, resourceLabel );
			strcpy ( tempOps.returnedResource, returnedResource );
			strcpy ( tempOps.specifiedResource, resource );
			tempOps.resourceType = ( unsigned int * ) 2;
			rfeObjectData.opOn.Add ( tempOps );
		}
		else
			return ( 0 );
	}

	return ( 1 );	// add successful target
}


int ResourceFileEditor::sysCall ( const char *command )
{	// code ripped from GNU 'system' man page
   int pid, status;

   if ( command == 0 )
       return 1;
	   
   pid = fork();
   
   if ( pid == -1 )
       return -1;
	   
   if ( pid == 0 )
   {
       char *argv[4];
       argv[0] = "sh";
       argv[1] = "-c";
       argv[2] = ( char * ) command;
       argv[3] = 0;
       execve( "/bin/sh", argv, environ );
       exit ( 127 );
   }

   do
   {
		if ( waitpid ( pid, &status, 0 ) == -1 )
		{
           if ( errno != EINTR )
               return -1;
		}
		else
           return status;   
   } while ( 1 );
}

